#=========================================================================
# StructuralRTLIRSignalExpr.py
#=========================================================================
"""RTLIR signal expression class definitions and generation method."""


import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits, is_bitstruct_inst
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt


class BaseSignalExpr:
  """Base abstract class of RTLIR signal expressions."""
  def __init__( s, rtype ):
    assert isinstance( rtype, rt.BaseRTLIRType ), f"non-RTLIR type {rtype} encountered!"
    s.rtype = rtype

  def get_rtype( s ):
    return s.rtype

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype

  def __ne__( s, other ):
    return not s.__eq__( other )

  def __hash__( s ):
    return hash((type(s), s.rtype))

#-------------------------------------------------------------------------
# Internal signal expression classes
#-------------------------------------------------------------------------

class _Index( BaseSignalExpr ):
  """Base IR class for all index signal operations.

  This class implements a series of static methods that is called to tell
  if the current index signal operation is indexing on a signal of packed
  data type array or a bit selection.
  """
  def __init__( s, index_base, index, rtype ):
    super().__init__( rtype )
    s.index = index
    s.base = index_base

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and\
           s.index == other.index and \
           s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.index, s.base))

  def get_index( s ):
    return s.index

  def get_base( s ):
    return s.base

  @staticmethod
  def is_packed_index( index_base, index ):
    """Return PackedIndex constructor if the given index base is a signal of packed data type."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, (rt.Port, rt.Wire)) and \
       isinstance(base_rtype.get_dtype(), rdt.PackedArray):
      return PackedIndex
    else:
      return None

  @staticmethod
  def is_bit_selection( index_base, index ):
    """Return BitSelection constructor if the given index base is a vector signal."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, (rt.Port, rt.Wire)) and \
       isinstance(base_rtype.get_dtype(), rdt.Vector):
      return BitSelection
    else:
      return None

class _UnpackedIndex( _Index ):
  """Base IR class for all index signal operations on unpacked arrays.

  Unpacked indexing operations include indexing on a port array, a wire
  array, a const array, an interface array, and a component array. This class
  implements a series of static methods to check if the given signal operation
  is one of them.
  """
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    rtype = base_rtype.get_next_dim_type()
    super().__init__( index_base, index, rtype )

  @staticmethod
  def is_port_index( index_base, index ):
    """Return PortIndex constructor if the given index base is a port array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Port):
      return PortIndex
    else:
      return None

  @staticmethod
  def is_wire_index( index_base, index ):
    """Return WireIndex constructor if the given index base is a wire array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Wire):
      return WireIndex
    else:
      return None

  @staticmethod
  def is_const_index( index_base, index ):
    """Return ConstIndex constructor if the given index base is a const array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Const):
      return ConstIndex
    else:
      return None

  @staticmethod
  def is_ifc_view_index( index_base, index ):
    """Return InterfaceViewIndex constructor if the given index base is an interface array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.InterfaceView):
      return InterfaceViewIndex
    else:
      return None

  @staticmethod
  def is_component_index( index_base, index ):
    """Return ComponentIndex constructor if the given index base is component array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Component):
      return ComponentIndex
    else:
      return None

class _Slice( BaseSignalExpr ):
  """Base IR class for all slicing signal operations.

  This class implements a single static method which checks if the given
  signal operation is slicing a vector signal (a.k.a. a part selection).
  """
  def __init__( s, slice_base, start, stop ):
    base_rtype = slice_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), rdt.Vector( stop-start ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( rdt.Vector( stop-start ) )
    else:
      assert False, f"unrecognized signal type {base_rtype} for slicing"
    super().__init__( rtype )
    s.base = slice_base
    s.slice = ( start, stop )

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and \
           s.slice == other.slice and s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.slice, s.base))

  def get_slice( s ):
    return s.slice

  def get_base( s ):
    return s.base

  @staticmethod
  def is_part_selection( slice_base, start, stop ):
    """Return PartSelection constructor if the given index base is a vector signal."""
    base_rtype = slice_base.get_rtype()
    if isinstance(base_rtype, (rt.Port, rt.Wire)) and \
       isinstance(base_rtype.get_dtype(), rdt.Vector) and \
       isinstance(start, int) and isinstance(stop, int):
      return PartSelection
    else:
      return None

class _Attribute( BaseSignalExpr ):
  """Base IR class for all attribute accessing signal operations.

  This class implements a series of static methods to check if the given signal
  operation is accessing attribute of the current component, a subcompoennt,
  an interface, or a field in a struct signal.
  """
  def __init__( s, attr_base, attr, rtype ):
    super().__init__( rtype )
    s.attr = attr
    s.base = attr_base

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and \
           s.attr == other.attr and \
           s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.attr, s.base))

  def get_base( s ):
    return s.base

  def get_attr( s ):
    return s.attr

  @staticmethod
  def is_cur_comp_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to the current component."""
    base_rtype = attr_base.get_rtype()
    if isinstance(attr_base, CurComp) and base_rtype.has_property(attr):
      return CurCompAttr
    else:
      return None

  @staticmethod
  def is_subcomp_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to a sub-component."""
    base_rtype = attr_base.get_rtype()
    if not isinstance(attr_base, CurComp) and \
       isinstance(base_rtype, rt.Component) and base_rtype.has_property(attr):
      return SubCompAttr
    else:
      return None

  @staticmethod
  def is_interface_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to an interface."""
    base_rtype = attr_base.get_rtype()
    if isinstance(base_rtype, rt.InterfaceView) and base_rtype.has_property(attr):
      return InterfaceAttr
    else:
      return None

  @staticmethod
  def is_struct_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to struct signal."""
    base_rtype = attr_base.get_rtype()
    if isinstance(base_rtype, rt.Signal):
      dtype = base_rtype.get_dtype()
      if isinstance(dtype, rdt.Struct) and dtype.has_property(attr):
        return StructAttr
      else:
        return None
    else:
      return None

#-------------------------------------------------------------------------
# Actual IR signal operation classes
#-------------------------------------------------------------------------

class ConstInstance( BaseSignalExpr ):
  """IR class for a constant instance.

  This class is for constants that appear in signal operations and is not
  an attribute of a component.
  """
  def __init__( s, obj, value ):
    super().__init__(rt.Const(rdt.get_rtlir_dtype( obj )))
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, ConstInstance) and s.rtype == other.rtype and \
           s.value == other.value

  def __hash__( s ):
    return hash((type(s), s.rtype, s.value))

  def get_value( s ):
    return s.value

class CurComp( BaseSignalExpr ):
  """IR class for the current component.

  This class is the base of most signal expressions because most signal
  expressions begin by accessing an attribute of the current component.
  This class implements a static method which checks if the given name
  is the same as the current component's name.
  """
  def __init__( s, comp, comp_id ):
    super().__init__(comp._pass_structural_rtlir_gen.rtlir_type)
    s.comp_id = comp_id

  def __eq__( s, other ):
    return isinstance(other, CurComp) and s.rtype == other.rtype and \
           s.comp_id == other.comp_id

  def __hash__( s ):
    return hash((type(s), s.rtype, s.comp_id))

  def get_component_id( s ):
    return s.comp_id

  @staticmethod
  def is_cur_comp( comp, comp_id ):
    """Return CurComp constructor if comp_id matches the current component's name."""
    if comp_id == comp._dsl.my_name:
      return CurComp
    else:
      return None

class PortIndex( _UnpackedIndex ):
  """IR class for indexing on a port array ."""

class WireIndex( _UnpackedIndex ):
  """IR class for indexing on a wire array ."""

class ConstIndex( _UnpackedIndex ):
  """IR class for indexing on a const array ."""

class InterfaceViewIndex( _UnpackedIndex ):
  """IR class for indexing on an interface array ."""

class ComponentIndex( _UnpackedIndex ):
  """IR class for indexing on a component array ."""

class PackedIndex( _Index ):
  """IR class for indexing on a signal of packed data type."""
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(),
                    dtype.get_next_dim_type() )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( dtype.get_next_dim_type() )
    else:
      assert False, f"unrecognized signal type {base_rtype} for indexing"
    super().__init__( index_base, index, rtype )

class BitSelection( _Index ):
  """IR class for selecting a bit of a vector signal."""
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), rdt.Vector( 1 ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( rdt.Vector( 1 ) )
    else:
      assert False, f"unrecognized signal type {base_rtype} for indexing"
    super().__init__( index_base, index, rtype )

class PartSelection( _Slice ):
  """IR class for selecting one or more bits of a vector signal."""

class CurCompAttr( _Attribute ):
  """IR class for accessing the attribute of the current component."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super().__init__( attr_base, attr, rtype )

class SubCompAttr( _Attribute ):
  """IR class for accessing the attribute of a sub-component."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super().__init__( attr_base, attr, rtype )

class InterfaceAttr( _Attribute ):
  """IR class for accessing the attribute of an interface."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super().__init__( attr_base, attr, rtype )

class StructAttr( _Attribute ):
  """IR class for accessing the attribute of a struct signal."""
  def __init__( s, attr_base, attr ):
    base_rtype = attr_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), dtype.get_property( attr ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( dtype.get_property( attr ) )
    else:
      assert False, f"unrecognized signal type {base_rtype} for field selection"
    super().__init__( attr_base, attr, rtype )

#-------------------------------------------------------------------------
# Map string signal expression to IR object generation methods
#-------------------------------------------------------------------------

signal_expr_classes = {
    'Attr' :  [ _Attribute.is_cur_comp_attr,   _Attribute.is_subcomp_attr,
                _Attribute.is_interface_attr,  _Attribute.is_struct_attr ],
    'Index' : [ _Index.is_packed_index,        _Index.is_bit_selection,
                _UnpackedIndex.is_port_index,  _UnpackedIndex.is_wire_index,
                _UnpackedIndex.is_const_index, _UnpackedIndex.is_ifc_view_index,
                _UnpackedIndex.is_component_index ],
    'Slice' : [ _Slice.is_part_selection ],
    'Base'  : [ CurComp.is_cur_comp ]
}

#-------------------------------------------------------------------------
# gen_signal_expr
#-------------------------------------------------------------------------

def gen_signal_expr( cur_component, signal ):
  """Return an RTLIR signal expression for the given signal."""

  def get_next_op( expr, cur_pos, my_name, full_name ):
    """Return the next signal operation in string `expr`.

    Return value: ( op, *data )
    op is one of [ 'Attr', 'Slice', 'Index', 'Base', 'Done' ].
    *data is one to two variables that describes the next signal operation.
    """
    if not expr[cur_pos:]: return ( 'Done', '' ), 0
    pos = len( expr )
    dot_pos = expr.find( '.', cur_pos+1 )
    lb_pos  = expr.find( '[', cur_pos+1 )
    pos = dot_pos if dot_pos != -1 and dot_pos < pos else pos
    pos =  lb_pos if  lb_pos != -1 and  lb_pos < pos else pos

    # Attribute operation
    if expr[cur_pos] == '.': return ( 'Attr', expr[cur_pos+1:pos] ), pos

    # Index or slice operation
    elif expr[cur_pos] == '[':
      rb_pos = expr.find( ']', cur_pos )
      colon_pos = expr.find( ':', cur_pos )
      assert rb_pos != -1 and pos == rb_pos+1, \
        f"unrecognized expression {expr}"

      if cur_pos < colon_pos < rb_pos:
        start = int( expr[cur_pos+1:colon_pos] )
        stop  = int( expr[colon_pos+1:rb_pos] )
        return ( 'Slice', start, stop ), pos
      else:
        return ( 'Index', int(expr[cur_pos+1:rb_pos]) ), pos

    # The current component ( base of attribute )
    else:
      base_pos = expr.find( full_name )
      assert base_pos >= 0, \
        f"cannot find the base of attribute {full_name} in {expr}"
      return ( 'Base', my_name ), base_pos + len( full_name )

  def get_cls_inst( func_list, cur_node, ops ):
    """Return an IR instance of the given signal operation."""
    classes = [ c for c in ( f( cur_node, *ops ) for f in func_list ) if c ]
    assert len(classes) <= 1, f"internal error: not unique class {classes}!"
    assert classes, f"internal error: no available expression nodes for {cur_node}!"
    return classes[0]( cur_node, *ops )

  try:

    try:
      expr = signal._dsl.full_name
      base_comp = signal
    except AttributeError:
      # Special case for a ConstInstance because it has no name
      assert hasattr( signal._dsl, 'const' ), f'{signal} is not supported!'
      c = signal._dsl.const
      assert isinstance( c, ( int, Bits )) or is_bitstruct_inst( c ), \
          f'{signal._dsl.const} is not an integer/Bits/BitStruct const!'
      return ConstInstance( signal, c )

    # Get the base component
    base_comp = cur_component
    full_name = base_comp._dsl.full_name
    my_name = base_comp._dsl.my_name
    assert expr.find( full_name ) >= 0, \
      f"cannot find the base of attribute {full_name} in {expr}"

    # Start from the base component and process one operation per iteration
    cur_pos, cur_node = 0, base_comp
    while cur_pos < len( expr ):
      op, next_pos = get_next_op( expr, cur_pos, my_name, full_name )
      if op[0] == 'Done': break
      cur_node = get_cls_inst( signal_expr_classes[ op[0] ], cur_node, op[1:] )
      cur_pos = next_pos
    return cur_node

  except AssertionError as e:
    msg = '' if e.args[0] is None else e.args[0]
    raise RTLIRConversionError( signal, msg )



#=========================================================================
# ImportConfigs.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jul 28, 2019
"""Configuration class of SystemVerilog import pass."""

import os
import subprocess
from textwrap import fill, indent

from pymtl3.dsl import Placeholder
from pymtl3.passes.errors import InvalidPassOptionValue
from pymtl3.passes.PassConfigs import BasePassConfigs, Checker
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir import get_component_ifc_rtlir

from ..util.utility import expand, get_component_unique_name


class ImportConfigs( BasePassConfigs ):

  Options = {
    # Import enable switch
    "import_" : True,

    # Enable verbose mode?
    "verbose" : False,

    # Port name mapping
    # Map pymtl names to Verilog names
    "port_map" : {},

    # Enable external line trace?
    # Once enabled, the `line_trace()` method of the imported component
    # will return a string read from the external `line_trace()` function.
    # This means your Verilog module has to have a `line_trace` function
    # that provides the line trace string which has less than 512 characters.
    # Default to False
    "external_trace" : False,

    # Verilator code generation options
    # These options will be passed to verilator to generate the C simulator.
    # By default, verilator is called with `--cc`.

    # --top-module
    # Expects the name of the top component;
    # "" to use name of the current component to be imported
    "top_module" : "",

    # Does the module to be imported has `clk` port?
    "has_clk" : True,

    # Does the module to be imported has `reset` port?
    "has_reset" : True,

    # Expects path of the file that contains the top module to be verilated
    # "" to use "<top_module>.sv"
    "vl_src" : "",

    # --Mdir
    # Expects the path of Makefile output directory;
    # "" to use `obj_dir_<top_module>`
    "vl_mk_dir" : "",

    # -f
    # Expects the path to the flist file; "" to disable this option
    "vl_flist" : "",

    # -I ( alias of -y and +incdir+ )
    # Expects a list of include paths; [] to disable this option
    "vl_include" : [],

    # --assert
    # Expects a boolean value
    "enable_assert" : True,

    # Verilator optimization options

    # -O0/3
    # Expects a non-negative integer
    # Currently only support 0 (disable opt) and 3 (highest effort opt)
    "vl_opt_level" : 3,

    # --unroll-count
    # Expects a non-negative integer
    # 0 to disable this option
    "vl_unroll_count" : 1000000,

    # --unroll-stmts
    # Expects a non-negative integer
    # 0 to disable this option
    "vl_unroll_stmts" : 1000000,

    # Verilator warning-related options

    # False to disable the warnings, True to enable
    "vl_W_lint" : True,
    "vl_W_style" : True,
    "vl_W_fatal" : True,

    # Un-warn all warnings in the given list; [] to disable this option
    # The given list should only include strings that appear in `Warnings`
    "vl_Wno_list" : [],

    # Verilator misc options

    # What is the inital value of signals?
    # Should be one of ['zeros', 'ones', 'rand']
    "vl_xinit" : "zeros",

    # --trace
    # Expects a boolean value
    "vl_trace" : False,

    # Passed to verilator tracing function
    "vl_trace_timescale" : "10ps",

    # `vl_trace_cycle_time`*`vl_trace_timescale` is the cycle time of the
    # PyMTL clock that appears in the generated VCD
    # With the default options, the frequency of PyMTL clock is 1GHz
    "vl_trace_cycle_time" : 100,

    # C-compilation options
    # These options will be passed to the C compiler to create a shared lib.

    # Additional flags to be passed to the C compiler.
    # By default, CC is called with `-O0 -fPIC -shared`.
    # "" to disable this option
    "c_flags" : "",

    # Additional include search path of the C compiler.
    # [] to disable this option
    "c_include_path" : [],

    # Additional C source files passed to the C compiler.
    # [] to compile verilator generated files only.
    "c_srcs" : [],

    # `LDLIBS` will be listed after the primary target file whereas
    # `LDFLAGS` will be listed before.

    # We enforce the GNU makefile implicit rule that `LDFLAGS` should only
    # include non-library linker flags such as `-L`.
    "ld_flags" : "",

    # We enforce the GNU makefile implicit rule that `LDLIBS` should only
    # include library linker flags/names such as `-lfoo`.
    "ld_libs" : "",

    # Enable all verilator coverage
    "coverage" : False,

    # Enable all verilator coverage
    "line_coverage" : False,

    # Enable all verilator coverage
    "toggle_coverage" : False,
  }

  Checkers = {
    ("import_", "enable_assert", "external_trace", "vl_W_lint", "vl_W_style",
     "vl_W_fatal", "vl_trace", "verbose", "has_clk", "has_reset", "coverage") :
      Checker( lambda v: isinstance(v, bool), "expects a boolean" ),

    ("c_flags", "ld_flags", "ld_libs") :
      Checker( lambda v: isinstance(v, str),  "expects a string" ),

    ("vl_unroll_count", "vl_unroll_stmts"):
      Checker( lambda v: isinstance(v, int) and v >= 0, "expects an integer >= 0" ),

    "top_module": Checker( lambda v: isinstance(v, str) and v, "expects a non-empty string"),

    "port_map": Checker( lambda v: isinstance(v, dict), "expects a dict"),

    "vl_src": Checker( lambda v: isinstance(v, str) and (os.path.isfile(expand(v)) or not v),
                "vl_src should be a path to a file or an empty string!" ),

    "vl_flist": Checker( lambda v: isinstance(v, str) and os.path.isfile(expand(v)) or v == "",
                         "expects a path to a file" ),

    "vl_Wno_list": Checker( lambda v: isinstance(v, list) and all(w in ImportConfigs.Warnings for w in v),
                            "expects a list of warnings" ),

    "vl_include": Checker( lambda v: isinstance(v, list) and all(os.path.isdir(expand(p)) for p in v),
                            "expects a path to directory"),

    "vl_xinit": Checker( lambda v: v in ['zeros', 'ones', 'rand'],
                  "vl_xinit should be one of ['zeros', 'ones', 'rand']" ),

    "vl_trace_timescale": Checker( lambda v: isinstance(v, str) and len(v) > 2 and v[-1] == 's' and \
                                    v[-2] in ['p', 'n', 'u', 'm'] and \
                                    all(c.isdigit() for c in v[:-2]),
                                    "expects a timescale string" ),

    "vl_trace_cycle_time": Checker( lambda v: isinstance(v, int) and (v % 2) == 0,
                                    "expects an integer `n` such that `n`*`vl_trace_timescale` is the cycle time" ),

    "vl_mk_dir": Checker( lambda v: isinstance(v, str), "expects a path to directory" ),

    "c_include_path": Checker( lambda v: isinstance(v, list) and all(os.path.isdir(expand(p)) for p in v),
                                "expects a list of paths to directories" ),

    "c_srcs": Checker( lambda v: isinstance(v, list) and all(os.path.isfile(expand(p)) for p in v),
                       "expects a list of paths to files" )
  }

  Warnings = [
    'ALWCOMBORDER', 'ASSIGNIN', 'ASSIGNDLY', 'BLKANDNBLK', 'BLKSEQ',
    'BLKLOOPINIT', 'BSSPACE', 'CASEINCOMPLETE', 'CASEOVERLAP',
    'CASEX', 'CASEWITHX', 'CDCRSTLOGIC', 'CLKDATA', 'CMPCONST',
    'COLONPLUS', 'COMBDLY', 'CONTASSREG', 'DECLFILENAME', 'DEFPARAM',
    'DETECTARRAY', 'ENDLABEL', 'GENCLK', 'IFDEPTH', 'IGNOREDRETURN',
    'IMPERFECTSCH', 'IMPLICIT', 'IMPORTSTAR', 'IMPURE', 'INCABSPATH',
    'INFINITELOOP', 'INITIALDLY', 'LITENDIAN', 'MODDUP', 'MULTIDRIVEN',
    'MULTITOP', 'PINCONNECTEMPTY', 'PINMISSING', 'PINNOCONNECT',
    'PROCASSWIRE', 'REALCVT', 'REDEFMACRO', 'SELRANGE', 'STMTDLY',
    'SYMRSVDWORD', 'SYNCASYNCNET', 'TASKNSVAR', 'TICKCOUNT', 'UNDRIVEN',
    'UNOPT', 'UNOPTFLAT', 'UNOPTTHREADS', 'UNPACKED', 'UNSIGNED', 'UNUSED',
    'USERINFO', 'USERWARN', 'USERERROR', 'USERFATAL', 'VARHIDDEN', 'WIDTH',
    'WIDTHCONCAT',
  ]

  PassName = 'sverilog.ImportPass'

  #-----------------------------------------------------------------------
  # Public APIs
  #-----------------------------------------------------------------------

  # Override
  def check( s ):
    super().check()
    
    if not s.vl_flist and not os.path.isfile(expand(s.vl_src)):
      raise InvalidPassOptionValue( 'vl_src', s.vl_src, s.PassName,
                'vl_src should be a path to a file when vl_flist is empty!' )

  def create_vl_cmd( s ):
    top_module  =  f"--top-module {s.top_module}"
    src         = s.vl_src
    mk_dir      = f"--Mdir {s.vl_mk_dir}"
    flist       = "" if s.is_default("vl_flist") else \
                  f"-f {s.vl_flist}"
    include     = "" if s.is_default("vl_include") else \
                  " ".join("-I" + path for path in s.vl_include)
    en_assert   = "--assert" if s.enable_assert else ""
    opt_level   = "-O3" if s.is_default( 'vl_opt_level' ) else "-O0"
    loop_unroll = "" if s.vl_unroll_count == 0 else \
                  f"--unroll-count {s.vl_unroll_count}"
    stmt_unroll = "" if s.vl_unroll_stmts == 0 else \
                  f"--unroll-stmts {s.vl_unroll_stmts}"
    trace       = "--trace" if s.vl_trace else ""
    coverage    = "--coverage" if s.coverage else ""
    line_cov    = "--coverage-line" if s.line_coverage else ""
    toggle_cov  = "--coverage-toggle" if s.toggle_coverage else ""
    warnings    = s.create_vl_warning_cmd()

    all_opts = [
      top_module, mk_dir, include, en_assert, opt_level, loop_unroll,
      stmt_unroll, trace, warnings, flist, src, coverage,
      line_cov, toggle_cov,
    ]

    return f"verilator --cc {' '.join(opt for opt in all_opts if opt)}"

  def create_cc_cmd( s ):
    c_flags = "-O0 -fPIC -fno-gnu-unique -shared" + \
             ("" if s.is_default("c_flags") else f" {expand(s.c_flags)}")
    c_include_path = " ".join("-I"+p for p in s.get_all_includes() if p)
    out_file = s.get_shared_lib_path()
    c_src_files = " ".join(s.get_c_src_files())
    ld_flags = expand(s.ld_flags)
    ld_libs = s.ld_libs
    coverage = "-DVM_COVERAGE" if s.coverage or \
                                  s.line_coverage or \
                                  s.toggle_coverage else ""
    return f"g++ {c_flags} {c_include_path} {ld_flags}"\
           f" -o {out_file} {c_src_files} {ld_libs} {coverage}"

  def fill_missing( s, m ):
    rtype = get_component_ifc_rtlir(m)
    s.v_param = rtype.get_params()
    s.is_param_wrapper = isinstance(m, Placeholder) and s.v_param
    top_module = s.top_module

    # Check if the keys of `port_map` are port names of `m`
    # Note that the keys can be expressions such as `ifc[0].foo` and
    # therefore we do not check if a port name of `m` is in the keys.
    if s.port_map:
      s.check_p_map( rtype )

    # Fill in the top module if unspecified
    # If the top-level module is not a wrapper
    if not s.is_param_wrapper:
      full_name = get_component_unique_name( rtype )
      if not top_module:
        # Default top_module is the name of component concatenated
        # with parameters
        s.top_module = full_name

    # If the top-level module is a wrapper
    else:
      # Check if all parameters are integers
      assert all(isinstance(v[1], int) for v in s.v_param), \
          "Only support integers as parameters of Verilog modules!"

      # wrapped_module is the name of the module to be parametrized
      # top_module is the name of the new top-level module
      if top_module:
        s.wrapped_module = top_module
        s.top_module = f"{top_module}_w_params"
      else:
        # Default top_module is the name of component
        s.wrapped_module = rtype.get_name()
        s.top_module = f"{s.wrapped_module}_w_params"

    top_module = s.top_module

    # Only try to infer the name of Verilog source file if both
    # flist and the source file are not specified.
    if not s.vl_src and not s.vl_flist:
      s.vl_src = f"{top_module}.sv"

    if not s.vl_mk_dir:
      s.vl_mk_dir = f"obj_dir_{top_module}"

    if s.is_param_wrapper:
      # If toplevel module is a param-wrapper, `vl_src` has to be specified
      # because the file containing the wrapper will include `vl_src` for
      # the module to be parametrized.
      if not s.vl_src:
        raise InvalidPassOptionValue(
            "vl_src", s.vl_src, s.PassName,
            "vl_src must be specified when Placeholder is to be imported!")
      s.v_module2param = s.vl_src
      s.vl_src = top_module+".sv"

  def is_top_wrapper( s ):
    return s.is_param_wrapper

  def is_port_mapped( s ):
    return bool(len(s.port_map) > 0)

  def get_port_map( s ):
    pmap = s.port_map
    return lambda name: pmap[name] if name in pmap else name

  def get_module_to_parametrize( s ):
    return s.wrapped_module

  def get_param_include( s ):
    return s.v_module2param

  def get_vl_xinit_value( s ):
    if s.vl_xinit == 'zeros':
      return 0
    elif s.vl_xinit == 'ones':
      return 1
    elif s.vl_xinit == 'rand':
      return 2
    else:
      raise InvalidPassOptionValue("vl_xinit should be one of 'zeros', 'ones', or 'rand'!")

  def get_c_wrapper_path( s ):
    return f"{s.top_module}_v.cpp"

  def get_py_wrapper_path( s ):
    return f"{s.top_module}_v.py"

  def get_shared_lib_path( s ):
    return f"lib{s.top_module}_v.so"

  def vprint( s, msg, nspaces = 0, use_fill = False ):
    if s.verbose:
      if use_fill:
        print(indent(fill(msg), " "*nspaces))
      else:
        print(indent(msg, " "*nspaces))

  #-----------------------------------------------------------------------
  # Internal helper methods
  #-----------------------------------------------------------------------

  def check_p_map( s, rtype ):
    """Check if each port name in the port map exists in component `rtype`."""
    pm = s.port_map
    all_ports = rtype.get_ports_packed()
    all_port_names = list(map(lambda x: x[0], rtype.get_ports_packed()))
    assert all(isinstance(p, rt.Port) and \
               isinstance(p.get_dtype(), rdt.Vector) for n, p in all_ports), \
        f"Port map option currently requires all ports of {rtype.get_name()}"\
        f" to be a single vector port."
    for name in pm.keys():
      if name not in all_port_names:
        raise InvalidPassOptionValue("port_map", pm, s.PassName,
          f"Port {name} does not exist in component {rtype.get_name()}!")

  def get_all_includes( s ):
    includes = s.c_include_path

    # Try to obtain verilator include path either from environment variable
    # or from `pkg-config`
    vl_include_dir = os.environ.get("PYMTL_VERILATOR_INCLUDE_DIR")
    if vl_include_dir is None:
      get_dir_cmd = ["pkg-config", "--variable=includedir", "verilator"]
      try:
        vl_include_dir = \
            subprocess.check_output(get_dir_cmd, stderr = subprocess.STDOUT).strip()
        vl_include_dir = vl_include_dir.decode('ascii')
      except OSError as e:
        vl_include_dir_msg = \
"""\
Cannot locate the include directory of verilator. Please make sure either \
$PYMTL_VERILATOR_INCLUDE_DIR is set or `pkg-config` has been configured properly!
"""
        raise OSError(fill(vl_include_dir_msg)) from e

    # Add verilator include path
    s.vl_include_dir = vl_include_dir
    includes += [vl_include_dir, vl_include_dir + "/vltstd"]

    return includes

  def get_c_src_files( s ):
    srcs = s.c_srcs
    top_module = s.top_module
    vl_mk_dir = s.vl_mk_dir
    vl_class_mk = f"{vl_mk_dir}/V{top_module}_classes.mk"

    # Add C wrapper
    srcs.append(s.get_c_wrapper_path())

    # Add files listed in class makefile
    with open(vl_class_mk, "r") as class_mk:
      srcs += s.get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_CLASSES_FAST")
      srcs += s.get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_CLASSES_SLOW")
      srcs += s.get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_SUPPORT_FAST")
      srcs += s.get_srcs_from_vl_class_mk(
          class_mk, vl_mk_dir, "VM_SUPPORT_SLOW")
      srcs += s.get_srcs_from_vl_class_mk(
          class_mk, s.vl_include_dir, "VM_GLOBAL_FAST")
      srcs += s.get_srcs_from_vl_class_mk(
          class_mk, s.vl_include_dir, "VM_GLOBAL_SLOW")

    return srcs

  def get_srcs_from_vl_class_mk( s, mk, path, label ):
    """Return all files under `path` directory in `label` section of `mk`."""
    srcs, found = [], False
    mk.seek(0)
    for line in mk:
      if line.startswith(label):
        found = True
      elif found:
        if line.strip() == "":
          found = False
        else:
          file_name = line.strip()[:-2]
          srcs.append( path + "/" + file_name + ".cpp" )
    return srcs

  def create_vl_warning_cmd( s ):
    lint = "" if s.is_default("vl_W_lint") else "--Wno-lint"
    style = "" if s.is_default("vl_W_style") else "--Wno-style"
    fatal = "" if s.is_default("vl_W_fatal") else "--Wno-fatal"
    wno = " ".join(f"--Wno-{w}" for w in s.vl_Wno_list)
    return " ".join(w for w in [lint, style, fatal, wno] if w)

  def is_default( s, opt ):
    return getattr( s, opt ) == s.Options[opt]



from pymtl3 import *
from pymtl3.stdlib.ifcs import MemMsgType
from pymtl3.stdlib.ifcs.mem_ifcs import MemMinionIfcFL

AMO_FUNS = { MemMsgType.AMO_ADD  : lambda m,a : m+a,
             MemMsgType.AMO_AND  : lambda m,a : m&a,
             MemMsgType.AMO_OR   : lambda m,a : m|a,
             MemMsgType.AMO_SWAP : lambda m,a : a,
             MemMsgType.AMO_MIN  : lambda m,a : m if m.int() < a.int() else a,
             MemMsgType.AMO_MINU : min,
             MemMsgType.AMO_MAX  : lambda m,a : m if m.int() > a.int() else a,
             MemMsgType.AMO_MAXU : max,
             MemMsgType.AMO_XOR  : lambda m,a : m^a,
           }

class MemoryFL( Component ):

  def construct( s, mem_nbytes=1<<20 ):
    s.mem = bytearray( mem_nbytes )

    s.ifc = MemMinionIfcFL( s.read, s.write, s.amo )

    s.trace = "     "
    @s.update
    def up_clear_trace():
      s.trace = "     "

  def read( s, addr, nbytes ):

    ret, shamt = Bits( nbytes << 3, 0 ), 0

    begin = int(addr)
    addr  = begin + nbytes - 1

    while addr >= begin:
      ret = (ret << 8) + s.mem[addr]
      addr -= 1
    s.trace = "[rd ]"
    return ret

  def write( s, addr, nbytes, data ):

    addr = int(addr)
    end  = addr + nbytes

    while addr < end:
      s.mem[addr] = data & 255
      data >>= 8
      addr += 1
    s.trace = "[wr ]"

  def amo( s, amo, addr, nbytes, data ):
    ret = s.read( addr, nbytes )
    s.write( addr, nbytes, AMO_FUNS[ int(amo) ]( ret, data ) )
    s.trace = "[amo]"
    return ret

  def read_mem( s, addr, size ):
    assert len(s.mem) > (addr + size)
    return s.mem[ addr : addr + size ]

  def write_mem( s, addr, data ):
    assert len(s.mem) > (addr + len(data))
    s.mem[ addr : addr + len(data) ] = data

  def line_trace( s ):
    return s.trace



from pymtl3.passes.backends.sverilog import ImportConfigs

from .import_.ImportPass import ImportPass
from .translation.TranslationPass import TranslationPass
from .TranslationImportPass import TranslationImportPass



from pymtl3 import *


class Reg( Component ):

  def construct( s, Type ):
    s.out = OutPort( Type )
    s.in_ = InPort( Type )

    @s.update_ff
    def up_reg():
      s.out <<= s.in_

  def line_trace( s ):
    return f"[{s.in_} > {s.out}]"

class RegEn( Component ):

  def construct( s, Type ):
    s.out = OutPort( Type )
    s.in_ = InPort( Type )

    s.en  = InPort( int if Type is int else Bits1 )

    @s.update_ff
    def up_regen():
      if s.en:
        s.out <<= s.in_

  def line_trace( s ):
    return f"[{'en' if s.en else '  '}|{s.in_} > {s.out}]"

class RegRst( Component ):

  def construct( s, Type, reset_value=0 ):
    s.out = OutPort( Type )
    s.in_ = InPort( Type )

    s.reset = InPort( int if Type is int else Bits1 )

    @s.update_ff
    def up_regrst():
      if s.reset: s.out <<= Type( reset_value )
      else:       s.out <<= s.in_

  def line_trace( s ):
    return f"[{'rst' if s.reset else '   '}|{s.in_} > {s.out}]"

class RegEnRst( Component ):

  def construct( s, Type, reset_value=0 ):
    s.out = OutPort( Type )
    s.in_ = InPort( Type )

    s.reset = InPort( int if Type is int else Bits1 )
    s.en    = InPort( int if Type is int else Bits1 )

    @s.update_ff
    def up_regenrst():
      if s.reset: s.out <<= Type( reset_value )
      elif s.en:  s.out <<= s.in_

  def line_trace( s ):
    return f"[{'en' if s.en else '  '}|{s.in_} > {s.out}]"



#=========================================================================
# SVBehavioralTranslatorL4.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 4 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL4 import (
    BehavioralTranslatorL4,
)
from pymtl3.passes.rtlir import RTLIRType as rt

from ...errors import SVerilogTranslationError
from .SVBehavioralTranslatorL3 import (
    BehavioralRTLIRToSVVisitorL3,
    SVBehavioralTranslatorL3,
)


class SVBehavioralTranslatorL4(
    SVBehavioralTranslatorL3, BehavioralTranslatorL4):

  def _get_rtlir2sv_visitor( s ):
    return BehavioralRTLIRToSVVisitorL4

#-------------------------------------------------------------------------
# BehavioralRTLIRToSVVisitorL4
#-------------------------------------------------------------------------

class BehavioralRTLIRToSVVisitorL4( BehavioralRTLIRToSVVisitorL3 ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for interface attributes in L4. We just name mangle every
    signal in an interface instead of constructing a SystemVerilog interface
    and instantiating new interfaces from it.
    """
    if isinstance( node.value.Type, rt.InterfaceView ):
      value = s.visit( node.value )
      attr = node.attr
      s.check_res( node, attr )
      return f'{value}__{attr}'
    else:
      return super().visit_Attribute( node )

  #-----------------------------------------------------------------------
  # visit_Index
  #-----------------------------------------------------------------------

  def visit_Index( s, node ):
    if isinstance( node.value.Type, rt.Array ) and \
        isinstance( node.value.Type.get_sub_type(), rt.InterfaceView ):
      try:
        nbits = node.idx._value
      except AttributeError:
        raise SVerilogTranslationError( s.blk, node,
          f'index of interface array {node.idx} must be a static constant expression!' )
      idx = int( nbits )
      value = s.visit( node.value )
      return f'{value}__{idx}'

    else:
      return super().visit_Index( node )



#=========================================================================
# TranslationImportPass.py
#=========================================================================
# Translate and import components having the `yosys_translate_import`
# attribute.
#
# Author : Peitian Pan
# Date   : Aug 6, 2019

from pymtl3.passes.backends.sverilog.TranslationImportPass import (
    TranslationImportPass as SVerilogTranslationImportPass,
)

from .import_.ImportPass import ImportPass
from .translation.TranslationPass import TranslationPass


class TranslationImportPass( SVerilogTranslationImportPass ):

  def get_translation_pass( s ):
    return TranslationPass()

  def get_import_pass( s ):
    return ImportPass()

  def get_flag_name( s ):
    return "yosys_translate_import"

  def get_translation_flag_name( s ):
    return "yosys_translate"

  def get_import_flag_name( s ):
    return "config_yosys_import"



#=========================================================================
# YosysStructuralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 13, 2019
"""Test the level 2 yosys-SystemVerilog structural translator."""

import pytest

from pymtl3.passes.backends.sverilog.util.test_utility import check_eq
from pymtl3.passes.backends.sverilog.util.utility import sverilog_reserved

from ....testcases import (
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectConstStructAttrToOutComp,
    CaseConnectLiteralStructComp,
    CaseConnectNestedStructPackedArrayComp,
)
from ..YosysStructuralTranslatorL2 import YosysStructuralTranslatorL2


def run_test( case, m ):
  m.elaborate()
  YosysStructuralTranslatorL2.is_sverilog_reserved = lambda s, x: x in sverilog_reserved
  tr = YosysStructuralTranslatorL2( m )
  tr.clear( m )
  tr.translate_structural( m )

  ports = tr.structural.decl_ports[m]
  wires = tr.structural.decl_wires[m]
  conns = tr.structural.connections[m]

  check_eq( ports["port_decls"],  case.REF_PORTS_PORT )
  check_eq( ports["wire_decls"],  case.REF_PORTS_WIRE )
  check_eq( ports["connections"], case.REF_PORTS_CONN )
  check_eq( wires, case.REF_WIRE )
  check_eq( conns, case.REF_CONN )

@pytest.mark.parametrize(
  'case', [
    CaseConnectConstStructAttrToOutComp,
    CaseConnectLiteralStructComp,
    CaseConnectArrayStructAttrToOutComp,
    CaseConnectNestedStructPackedArrayComp,
  ]
)
def test_yosys_structural_L2( case ):
  run_test( case, case.DUT() )



#=========================================================================
# BehavioralRTLIRL1Pass_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Feb 2, 2019
"""Test the level 1 behavioral RTLIR passes.

The L1 generation, L1 type check, and visualization passes are invoked. The
generation pass results are verified against a reference AST.
"""

import sys

import pytest

from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL1Pass import (
    BehavioralRTLIRGenL1Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL1Pass import (
    BehavioralRTLIRTypeCheckL1Pass,
)
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError, PyMTLTypeError
from pymtl3.passes.rtlir.util.test_utility import do_test, expected_failure
from pymtl3.passes.testcases import (
    CaseAssignMultiTargetComp,
    CaseAttributeSignalComp,
    CaseBitsArgsComp,
    CaseBitSelOutOfRangeComp,
    CaseClassdefComp,
    CaseComponentBaseIndexComp,
    CaseComponentHigherSliceComp,
    CaseComponentInIndexComp,
    CaseComponentLowerSliceComp,
    CaseConcatComponentComp,
    CaseConcatNoArgsComp,
    CaseCopyArgsComp,
    CaseDeepcopyArgsComp,
    CaseDeleteComp,
    CaseDictComp,
    CaseDictComprehensionComp,
    CaseDoubleStarArgsComp,
    CaseDroppedAttributeComp,
    CaseExecComp,
    CaseExtendedSubscriptComp,
    CaseExtSliceComp,
    CaseFuncNotFoundComp,
    CaseGeneratorExprComp,
    CaseGlobalComp,
    CaseImportComp,
    CaseImportFromComp,
    CaseIndexOnStructComp,
    CaseIndexOutOfRangeComp,
    CaseKwArgsComp,
    CaseL1UnsupportedSubCompAttrComp,
    CaseLambdaFuncComp,
    CaseLHSComponentComp,
    CaseLHSConstComp,
    CaseListComp,
    CaseListComprehensionComp,
    CaseNonNameCalledComp,
    CasePassComp,
    CaseRaiseComp,
    CaseReprComp,
    CaseRHSComponentComp,
    CaseSetComp,
    CaseSetComprehensionComp,
    CaseSextOnComponentComp,
    CaseSextSmallNbitsComp,
    CaseSextTwoArgsComp,
    CaseSextVaribleNbitsComp,
    CaseSizeCastComponentComp,
    CaseSliceBoundLowerComp,
    CaseSliceOnComponentComp,
    CaseSliceOnStructComp,
    CaseSliceOutOfRangeComp,
    CaseSliceVariableBoundComp,
    CaseSliceWithStepsComp,
    CaseStandaloneExprComp,
    CaseStarArgsComp,
    CaseStrComp,
    CaseStructBitsUnagreeableComp,
    CaseTmpComponentComp,
    CaseTryExceptComp,
    CaseTryFinallyComp,
    CaseTupleComp,
    CaseUnrecognizedFuncComp,
    CaseUntypedTmpComp,
    CaseUpblkArgComp,
    CaseWhileComp,
    CaseWithComp,
    CaseYieldComp,
    CaseZextOnComponentComp,
    CaseZextSmallNbitsComp,
    CaseZextTwoArgsComp,
    CaseZextVaribleNbitsComp,
)

# TODO: fix all tests with this mark, including in other files (grep!).
# (we use the mark instead of allowing failure more generally because this
#  prevents regressions in all the other tests that *pass* on Python 3)
# If a test passes despite the mark, just remove it from that test!
if sys.version_info[0] == 3:
  XFAIL_ON_PY3 = pytest.mark.xfail(strict=True)
else:
  XFAIL_ON_PY3 = lambda f: f


def local_do_test( m ):
  """Check if generated behavioral RTLIR is the same as reference."""
  m = m.DUT()
  m.elaborate()
  m.apply( BehavioralRTLIRGenL1Pass() )
  m.apply( BehavioralRTLIRTypeCheckL1Pass() )

  try:
    ref = m._rtlir_test_ref
    for blk in m.get_update_blocks():
      upblk = m._pass_behavioral_rtlir_gen.rtlir_upblks[ blk ]
      assert upblk == ref[ blk.__name__ ]
  except AttributeError:
    pass

#-------------------------------------------------------------------------
# PyMTL type errors
#-------------------------------------------------------------------------

def test_L1_assign_unagreeable( do_test ):
  with expected_failure( PyMTLTypeError, "Unagreeable types" ):
    do_test( CaseStructBitsUnagreeableComp )

def test_L1_concat_component( do_test ):
  with expected_failure( PyMTLTypeError, "not a signal" ):
    do_test( CaseConcatComponentComp )

def test_L1_zext_variable_nbits( do_test ):
  with expected_failure( PyMTLTypeError, "not a constant number" ):
    do_test( CaseZextVaribleNbitsComp )

def test_L1_zext_small_nbits( do_test ):
  with expected_failure( PyMTLTypeError, "not greater" ):
    do_test( CaseZextSmallNbitsComp )

def test_L1_sext_variable_nbits( do_test ):
  with expected_failure( PyMTLTypeError, "not a constant number" ):
    do_test( CaseSextVaribleNbitsComp )

def test_L1_sext_small_nbits( do_test ):
  with expected_failure( PyMTLTypeError, "not greater" ):
    do_test( CaseSextSmallNbitsComp )

def test_L1_wrong_attribute( do_test ):
  with expected_failure( PyMTLTypeError, "does not have attribute in_" ):
    do_test( CaseDroppedAttributeComp )

def test_L1_unsupported_attr( do_test ):
  with expected_failure( PyMTLTypeError, "not supported at L1" ):
    do_test( CaseL1UnsupportedSubCompAttrComp )

def test_L1_array_index_out_of_range( do_test ):
  with expected_failure( PyMTLTypeError, "array index out of range" ):
    do_test( CaseIndexOutOfRangeComp )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_bit_sel_index_out_of_range( do_test ):
  with expected_failure( PyMTLTypeError, "bit selection index out of range" ):
    do_test( CaseBitSelOutOfRangeComp )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_index_on_struct( do_test ):
  with expected_failure( PyMTLTypeError, "cannot perform index on Struct" ):
    do_test( CaseIndexOnStructComp )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_slice_on_struct( do_test ):
  with expected_failure( PyMTLTypeError, "cannot perform slicing on type Struct" ):
    do_test( CaseSliceOnStructComp )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_slice_l_ngt_r( do_test ):
  with expected_failure( PyMTLTypeError, "must be larger than the lower bound" ):
    do_test( CaseSliceBoundLowerComp )

def test_L1_slice_variable_bound( do_test ):
  with expected_failure( PyMTLTypeError, "slice bounds must be constant" ):
    do_test( CaseSliceVariableBoundComp )

@pytest.mark.xfail( reason = "PyMTL DSL already threw InvalidIndexError" )
def test_L1_slice_out_of_range( do_test ):
  with expected_failure( PyMTLTypeError, "slice out of width of signal" ):
    do_test( CaseSliceOutOfRangeComp )

#-------------------------------------------------------------------------
# PyMTL type errors: unexpected types
#-------------------------------------------------------------------------

def test_L1_assign_lhs_const( do_test ):
  with expected_failure( PyMTLTypeError, "must be a signal" ):
    do_test( CaseLHSConstComp )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_L1_assign_lhs_component( do_test ):
  with expected_failure( PyMTLTypeError, "must be a signal" ):
    do_test( CaseLHSComponentComp )

def test_L1_assign_rhs_component( do_test ):
  with expected_failure( PyMTLTypeError, "signal or const" ):
    do_test( CaseRHSComponentComp )

def test_L1_zext_component( do_test ):
  with expected_failure( PyMTLTypeError, "only applies to signals" ):
    do_test( CaseZextOnComponentComp )

def test_L1_sext_component( do_test ):
  with expected_failure( PyMTLTypeError, "only applies to signals" ):
    do_test( CaseSextOnComponentComp )

def test_L1_size_cast_component( do_test ):
  with expected_failure( PyMTLTypeError, "size casting only applies to signals/consts" ):
    do_test( CaseSizeCastComponentComp )

def test_L1_attr_signal( do_test ):
  with expected_failure( PyMTLTypeError, "base of an attribute must be a component" ):
    do_test( CaseAttributeSignalComp )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_L1_index_component( do_test ):
  with expected_failure( PyMTLTypeError, "signal or constant expression" ):
    do_test( CaseComponentInIndexComp )

def test_L1_index_base_component( do_test ):
  with expected_failure( PyMTLTypeError, "base of an index must be an array or signal" ):
    do_test( CaseComponentBaseIndexComp )

def test_L1_slice_lower_component( do_test ):
  with expected_failure( PyMTLTypeError, "constant expression" ):
    do_test( CaseComponentLowerSliceComp )

def test_L1_slice_upper_component( do_test ):
  with expected_failure( PyMTLTypeError, "constant expression" ):
    do_test( CaseComponentHigherSliceComp )

def test_L1_slice_component( do_test ):
  with expected_failure( PyMTLTypeError, "base of a slice must be a signal" ):
    do_test( CaseSliceOnComponentComp )

#-------------------------------------------------------------------------
# PyMTL syntax errors
#-------------------------------------------------------------------------

def test_L1_upblk_arg( do_test ):
  with expected_failure( PyMTLSyntaxError, "not have arguments" ):
    do_test( CaseUpblkArgComp )

def test_L1_multi_assign( do_test ):
  with expected_failure( PyMTLSyntaxError, "multiple targets" ):
    do_test( CaseAssignMultiTargetComp )

def test_L1_copy_arg( do_test ):
  with expected_failure( PyMTLSyntaxError, "takes exactly 1 argument" ):
    do_test( CaseCopyArgsComp )

def test_L1_deepcopy_arg( do_test ):
  with expected_failure( PyMTLSyntaxError, "takes exactly 1 argument" ):
    do_test( CaseDeepcopyArgsComp )

def test_L1_slice_step( do_test ):
  with expected_failure( PyMTLSyntaxError, "Slice with steps" ):
    do_test( CaseSliceWithStepsComp )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_L1_illegal_subscript( do_test ):
  with expected_failure( PyMTLSyntaxError, "Illegal subscript" ):
    do_test( CaseExtendedSubscriptComp )

def test_L1_closure_freevar( do_test ):
  with expected_failure( PyMTLSyntaxError, "not a sub-component" ):
    do_test( CaseTmpComponentComp )

def test_L1_tmpvar( do_test ):
  with expected_failure( PyMTLSyntaxError, "Temporary variable" ):
    do_test( CaseUntypedTmpComp )

#-------------------------------------------------------------------------
# Call-related syntax errors
#-------------------------------------------------------------------------

@XFAIL_ON_PY3
def test_L1_call_star_arg( do_test ):
  with expected_failure( PyMTLSyntaxError, "star argument" ):
    do_test( CaseStarArgsComp )

@XFAIL_ON_PY3
def test_L1_call_double_star_arg( do_test ):
  with expected_failure( PyMTLSyntaxError, "double-star argument" ):
    do_test( CaseDoubleStarArgsComp )

def test_L1_call_keyword_arg( do_test ):
  with expected_failure( PyMTLSyntaxError, "keyword argument" ):
    do_test( CaseKwArgsComp )

def test_L1_call_name( do_test ):
  with expected_failure( PyMTLSyntaxError, "called but is not a name" ):
    do_test( CaseNonNameCalledComp )

def test_L1_call_not_found( do_test ):
  with expected_failure( PyMTLSyntaxError, "function is not found" ):
    do_test( CaseFuncNotFoundComp )

def test_L1_call_Bits_args( do_test ):
  with expected_failure( PyMTLSyntaxError, "exactly one argument" ):
    do_test( CaseBitsArgsComp )

def test_L1_call_concat_args( do_test ):
  with expected_failure( PyMTLSyntaxError, "at least one argument" ):
    do_test( CaseConcatNoArgsComp )

def test_L1_call_zext_num_args( do_test ):
  with expected_failure( PyMTLSyntaxError, "exactly two arguments" ):
    do_test( CaseZextTwoArgsComp )

def test_L1_call_sext_num_args( do_test ):
  with expected_failure( PyMTLSyntaxError, "exactly two arguments" ):
    do_test( CaseSextTwoArgsComp )

def test_L1_call_unrecognized( do_test ):
  with expected_failure( PyMTLSyntaxError, "Unrecognized method" ):
    do_test( CaseUnrecognizedFuncComp )

#-------------------------------------------------------------------------
# PyMTL syntax errors for unsupported Python syntax
#-------------------------------------------------------------------------

def test_Expr( do_test ):
  with expected_failure( PyMTLSyntaxError, "Stand-alone expression" ):
    do_test( CaseStandaloneExprComp )

def test_Lambda( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: lambda function" ):
    do_test( CaseLambdaFuncComp )

def test_Dict( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid type: dict" ):
    do_test( CaseDictComp )

def test_Set( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid type: set" ):
    do_test( CaseSetComp )

def test_List( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid type: list" ):
    do_test( CaseListComp )

def test_Tuple( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid type: tuple" ):
    do_test( CaseTupleComp )

def test_ListComp( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: list comprehension" ):
    do_test( CaseListComprehensionComp )

def test_SetComp( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: set comprehension" ):
    do_test( CaseSetComprehensionComp )

def test_DictComp( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: dict comprehension" ):
    do_test( CaseDictComprehensionComp )

def test_GeneratorExp( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: generator expression" ):
    do_test( CaseGeneratorExprComp )

def test_Yield( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: yield" ):
    do_test( CaseYieldComp )

@pytest.mark.xfail( reason = "Syntax not compatible with Python 3" )
def test_Repr( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: repr" ):
    do_test( CaseReprComp )

def test_Str( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: str" ):
    do_test( CaseStrComp )

def test_Classdef( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: classdef" ):
    do_test( CaseClassdefComp )

def test_Delete( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: delete" ):
    do_test( CaseDeleteComp )

def test_With( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: with" ):
    do_test( CaseWithComp )

def test_Raise( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: raise" ):
    do_test( CaseRaiseComp )

@XFAIL_ON_PY3
def test_TryExcept( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: try-except" ):
    do_test( CaseTryExceptComp )

@XFAIL_ON_PY3
def test_TryFinally( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: try-finally" ):
    do_test( CaseTryFinallyComp )

def test_Import( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: import" ):
    do_test( CaseImportComp )

def test_ImportFrom( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: import-from" ):
    do_test( CaseImportFromComp )

@pytest.mark.xfail( reason = "Syntax not compatible with Python 3" )
def test_Exec( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: exec" ):
    do_test( CaseExecComp )

def test_Global( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: global" ):
    do_test( CaseGlobalComp )

def test_Pass( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: pass" ):
    do_test( CasePassComp )

def test_While( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: while" ):
    do_test( CaseWhileComp )

@pytest.mark.xfail( reason = "PyMTL DSL AST parsing failed" )
def test_ExtSlice( do_test ):
  with expected_failure( PyMTLSyntaxError, "invalid operation: extslice" ):
    do_test( CaseExtSliceComp )



"""Expose method that returns RTLIR.

This module exposes get_rtlir method that converts PyMTL components
to RTLIR representation.
"""
from .behavioral import (
    BehavioralRTLIR,
    BehavioralRTLIRGenPass,
    BehavioralRTLIRTypeCheckPass,
    BehavioralRTLIRVisualizationPass,
)
from .rtype import RTLIRDataType, RTLIRType
from .rtype.RTLIRDataType import get_rtlir_dtype
from .rtype.RTLIRType import get_component_ifc_rtlir, get_rtlir, is_rtlir_convertible
from .structural import StructuralRTLIRGenPass, StructuralRTLIRSignalExpr



#=========================================================================
# SVBehavioralTranslatorL2.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 2 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL2 import (
    BehavioralTranslatorL2,
)
from pymtl3.passes.rtlir import BehavioralRTLIR as bir
from pymtl3.passes.rtlir import RTLIRType as rt

from ...util.utility import make_indent
from .SVBehavioralTranslatorL1 import (
    BehavioralRTLIRToSVVisitorL1,
    SVBehavioralTranslatorL1,
)


class SVBehavioralTranslatorL2( SVBehavioralTranslatorL1, BehavioralTranslatorL2 ):

  def _get_rtlir2sv_visitor( s ):
    return BehavioralRTLIRToSVVisitorL2

  def rtlir_tr_behavioral_tmpvars( s, tmpvars ):
    make_indent( tmpvars, 1 )
    return '\n'.join( tmpvars )

  def rtlir_tr_behavioral_tmpvar( s, id_, upblk_id, dtype ):
    return s.rtlir_tr_wire_decl(
        "__tmpvar__"+upblk_id+'_'+id_, rt.Wire(dtype['raw_dtype']),
        s.rtlir_tr_unpacked_array_type(None), dtype )

#-------------------------------------------------------------------------
# BehavioralRTLIRToSVVisitorL2
#-------------------------------------------------------------------------

class BehavioralRTLIRToSVVisitorL2( BehavioralRTLIRToSVVisitorL1 ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  def __init__( s, is_reserved ):
    super().__init__( is_reserved )

    # The dictionary of operator-character pairs
    s.ops = {
      # Unary operators
      bir.Invert : '~', bir.Not : '!', bir.UAdd : '+', bir.USub : '-',
      # Boolean operators
      bir.And : '&&', bir.Or : '||',
      # Binary operators
      bir.Add : '+', bir.Sub : '-', bir.Mult : '*', bir.Div : '/',
      bir.Mod : '%', bir.Pow : '**',
      bir.ShiftLeft : '<<', bir.ShiftRightLogic : '>>',
      bir.BitAnd : '&', bir.BitOr : '|', bir.BitXor : '^',
      # Comparison operators
      bir.Eq : '==', bir.NotEq : '!=', bir.Lt : '<', bir.LtE : '<=',
      bir.Gt : '>', bir.GtE : '>='
    }

  def visit_expr_wrap( s, node ):
    """Return expressions selectively wrapped with brackets."""
    if isinstance( node,
        ( bir.IfExp, bir.UnaryOp, bir.BoolOp, bir.BinOp, bir.Compare ) ):
      return f"( {s.visit(node)} )"
    else:
      return s.visit( node )

  #-----------------------------------------------------------------------
  # Statements
  #-----------------------------------------------------------------------
  # All statement nodes return a list of strings.

  #-----------------------------------------------------------------------
  # visit_If
  #-----------------------------------------------------------------------

  def visit_If( s, node ):
    src    = []
    body   = []
    orelse = []

    # Grab condition, if-body, and orelse-body
    cond = s.visit( node.cond )
    for stmt in node.body:
      body.extend( s.visit( stmt ) )
    make_indent( body, 1 )
    for stmt in node.orelse:
      orelse.extend( s.visit( stmt ) )

    # Assemble the statement, starting with if-body
    if_begin   = f'if ( {cond} ) begin'
    src.extend( [ if_begin ] )
    src.extend( body )

    # if len( node.body ) > 1:
    src.extend( [ 'end' ] )

    # If orelse-body is not empty, add it to the list of strings
    if node.orelse != []:

      # If an if statement is the only statement in the orelse-body
      if len( node.orelse ) == 1 and isinstance( node.orelse[ 0 ], bir.If ):
        # No indent will be added, also append if-begin to else-begin
        else_begin = f'else {orelse[0]}'
        orelse = orelse[ 1 : ]

      # Else indent orelse-body
      else:
        else_begin = 'else' + ( ' begin' if len( node.orelse ) > 1 else '' )
        make_indent( orelse, 1 )

      src.extend( [ else_begin ] )
      src.extend( orelse )
      if len( node.orelse ) > 1:
        src.extend( [ 'end' ] )

    return src

  #-----------------------------------------------------------------------
  # visit_For
  #-----------------------------------------------------------------------

  def visit_For( s, node ):
    src      = []
    body     = []
    loop_var = s.visit( node.var )
    start    = s.visit( node.start )
    end      = s.visit( node.end )

    begin    = ' begin' if len( node.body ) > 1 else ''

    cmp_op   = '<' if node.step.value > 0 else '<'
    inc_op   = '+' if node.step.value > 0 else '-'

    step_abs = s.visit( node.step )
    step_abs = step_abs if node.step.value > 0 else step_abs[ 1 : ]

    for stmt in node.body:
      body.extend( s.visit( stmt ) )
    make_indent( body, 1 )

    for_begin = \
      'for ( int {v} = {s}; {v} {comp} {t}; {v} {inc}= {stp} ){begin}'.format(
      v = loop_var, s = start, t = end, stp = step_abs,
      comp = cmp_op, inc = inc_op, begin = begin
    )

    # Assemble for statement
    src.extend( [ for_begin ] )
    src.extend( body )

    if len( node.body ) > 1:
      src.extend( [ 'end' ] )

    return src

  #-----------------------------------------------------------------------
  # Expressions
  #-----------------------------------------------------------------------
  # All expression nodes return a single string.

  #-----------------------------------------------------------------------
  # visit_IfExp
  #-----------------------------------------------------------------------

  def visit_IfExp( s, node ):
    cond  = s.visit_expr_wrap( node.cond )
    true  = s.visit( node.body )
    false = s.visit( node.orelse )

    return f'{cond} ? {true} : {false}'

  #-----------------------------------------------------------------------
  # visit_UnaryOp
  #-----------------------------------------------------------------------

  def visit_UnaryOp( s, node ):
    op      = s.ops[ type( node.op ) ]
    operand = s.visit_expr_wrap( node.operand )

    return f'{op}{operand}'

  #-----------------------------------------------------------------------
  # visit_BoolOp
  #-----------------------------------------------------------------------

  def visit_BoolOp( s, node ):
    op     = s.ops[ type( node.op ) ]
    values = []

    for value in node.values:
      values.append( s.visit_expr_wrap( value ) )
    src = f' {op} '.join( values )

    return src

  #-----------------------------------------------------------------------
  # visit_BinOp
  #-----------------------------------------------------------------------

  def visit_BinOp( s, node ):
    op  = s.ops[ type( node.op ) ]
    lhs = s.visit_expr_wrap( node.left )
    rhs = s.visit_expr_wrap( node.right )

    return f'{lhs} {op} {rhs}'

  #-----------------------------------------------------------------------
  # visit_Compare
  #-----------------------------------------------------------------------

  def visit_Compare( s, node ):
    op  = s.ops[ type( node.op ) ]
    lhs = s.visit_expr_wrap( node.left )
    rhs = s.visit_expr_wrap( node.right )

    return f'{lhs} {op} {rhs}'

  #-----------------------------------------------------------------------
  # visit_LoopVar
  #-----------------------------------------------------------------------

  def visit_LoopVar( s, node ):
    s.check_res( node, node.name )
    return node.name

  #-----------------------------------------------------------------------
  # visit_TmpVar
  #-----------------------------------------------------------------------

  def visit_TmpVar( s, node ):
    return f"__tmpvar__{node.upblk_name}_{node.name}"

  #-----------------------------------------------------------------------
  # visit_LoopVarDecl
  #-----------------------------------------------------------------------

  def visit_LoopVarDecl( s, node ):
    s.check_res( node, node.name )
    return node.name



#=========================================================================
# StructuralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 21, 2019
"""Test the level 3 structural translators."""

import pytest

from ...testcases import (
    CaseArrayBits32IfcInComp,
    CaseConnectArrayNestedIfcComp,
    CaseConnectValRdyIfcComp,
)
from ..StructuralTranslatorL3 import StructuralTranslatorL3
from .TestStructuralTranslator import mk_TestStructuralTranslator


def run_test( case, m ):
  m.elaborate()
  tr = mk_TestStructuralTranslator(StructuralTranslatorL3)(m)
  tr.clear( m )
  tr.translate_structural(m)
  try:
    decl_ifcs = tr.structural.decl_ifcs[m]
    assert decl_ifcs == case.REF_IFC
    connections = tr.structural.connections[m]
    assert connections == case.REF_CONN
  except AttributeError:
    pass

@pytest.mark.parametrize(
  'case', [
    CaseConnectValRdyIfcComp,
    CaseArrayBits32IfcInComp,
    CaseConnectArrayNestedIfcComp,
  ]
)
def test_generic_structural_L3( case ):
  run_test( case, case.DUT() )



"""
========================================================================
SimpleTickPass.py
========================================================================
Generate a simple tick function (no Mamba techniques here)

Author : Shunning Jiang
Date   : Dec 26, 2018
"""
from pymtl3.dsl import MethodPort
from pymtl3.dsl.errors import UpblkCyclicError
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError

# Shunning: We are aware of the problem that there may be multiple places
# that assembles the function for the user to tick the simulator.
# I think LLVM applies passes based on the order they are inserted to the
# pass manager. LLVM also has some local ordering mechanism to figure out
# the dependencies between passes. Currently we haven't got to the point
# where we have enough passes to manage, but we should keep this in mind.

class SimpleTickPass( BasePass ):

  @staticmethod
  def gen_tick_function( schedule ):
    def iterative():
      for blk in schedule:
        blk()
    return iterative

  def __call__( self, top ):
    if not hasattr( top._sched, "update_schedule" ):
      raise PassOrderError( "update_schedule" )

    # We assemble the final schedule from multiple sources of required
    # work to generate a tick function for simulation

    # Currently the tick order is:
    # [ ff, tracing, posedge, clear_cl_trace, update ]

    final_schedule = []
    # call ff blocks first
    final_schedule.extend( top._sched.schedule_ff )

    # append tracing related work

    if hasattr( top, "_tracing" ):
      if hasattr( top._tracing, "vcd_func" ):
        final_schedule.append( top._tracing.vcd_func )
      if hasattr( top._tracing, "collect_text_sigs" ):
        final_schedule.append( top._tracing.collect_text_sigs )

    # posedge flip
    final_schedule.extend( top._sched.schedule_posedge_flip )

    # advance cycle after posedge
    def generate_advance_sim_cycle( top ):
      def advance_sim_cycle():
        top.simulated_cycles += 1
      return advance_sim_cycle
    final_schedule.append( generate_advance_sim_cycle(top) )

    # clear cl method flag
    if hasattr( top, "_tracing" ):
      if hasattr( top._tracing, "clear_cl_trace" ):
        final_schedule.append( top._tracing.clear_cl_trace )

    # execute all update blocks
    final_schedule.extend( top._sched.update_schedule )

    # Generate tick
    top.tick = SimpleTickPass.gen_tick_function( final_schedule )
    # reset sim_cycles
    top.simulated_cycles = 0

    # FIXME update_once?
    # check if the design has method_port
    method_ports = top.get_all_object_filter( lambda x: isinstance( x, MethodPort ) )

    if len(method_ports) == 0:
      # Pure RTL design, add eval_combinational
      top.eval_combinational = SimpleTickPass.gen_tick_function( top._sched.update_schedule )
    else:
      tmp = list(method_ports)[0]
      def eval_combinational():
        raise NotImplementedError(f"top is not a pure RTL design. {'top'+repr(tmp)[1:]} is a method port.")
      top.eval_combinational = eval_combinational



#=========================================================================
# PassConfigs.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jul 26, 2019

from copy import deepcopy

from .errors import InvalidPassOption, InvalidPassOptionValue


class Checker:

  def __init__( s, condition=lambda x: True, error_msg="" ):
    s.condition = condition
    s.error_msg = error_msg

class BasePassConfigs:
  """Base class of customized pass configrations.

  This base class provides methods to get/set option values, set the checking
  functions (checkers) for one or more options, and check if all given options
  are valid. Subclasses will provide all possible options with their default
  values. After initialization, all options and their values can be accessed
  from `s.options` or through `get_option` and `set_option`.

  Users who wish to use this Base class should (1) put the name of their
  pass to `PassName`; (2) fill their own options and default values into
  `Options` class attribute; (3) call `set_checker(s)` to define checkers for
  options or manually define `check_<option_name>` methods that takes the value
  of `<option_name>` and returns whether this option is valid. Note that this
  base class assumes a method with name `check_<option_name>` is a checker and
  will call that method to determine whether the value of option `<option_name>`
  is valid.
  """
  def __new__( cls, *args, **kwargs ):
    assert len(args) == 0, "We only accept keyword arguments here."
    assert hasattr( cls, "Options"  )
    assert hasattr( cls, "PassName" )
    assert hasattr( cls, "Checkers" )

    inst = super().__new__( cls )

    opts = deepcopy( cls.Options )
    for opt, value in kwargs.items():
      if opt not in cls.Options:
        raise InvalidPassOption( opt, inst.PassName )
      opts[opt] = value

    cls.opts = opts.keys()

    # Preprocess checkers
    cls._Checkers = {}
    for opt, chk in cls.Checkers.items():
      assert isinstance( chk, Checker ), f'Checker for "{opt}" can only be an instance of Checker, not {chk}.'

      if isinstance( opt, tuple ):
        for op in opt:
          assert op in opts, f"'{op}' is not a valid operation so we cannot set Checker for it."
          cls._Checkers[ op ] = chk

      elif isinstance( opt, str ):
        assert opt in opts, f"'{op}' is not a valid operation so we cannot set Checker for it."
        cls._Checkers[ opt ] = chk

      else:
        raise InvalidPassOption(f"Option name can only be a tuple of strings (a,b,c) or string a, not '{op}'")

    trivial_checker = Checker( lambda x: True, "" )
    for opt, value in opts.items():
      if opt not in cls._Checkers:
        cls._Checkers[opt] = trivial_checker
      assert not hasattr( inst, opt ), f"There is already a field in config called '{opt}'. What happened?"
      setattr( inst, opt, value )

    return inst

  def check( s ):
    """Check whether the given options are valid by calling checkers."""
    for opt in s.opts:
      chk   = s._Checkers[opt]
      value = getattr( s, opt )
      if not chk.condition( value ):
        raise InvalidPassOptionValue( opt, value, s.PassName, chk.error_msg )



from pymtl3.dsl import Component

from .autotick.OpenLoopCLPass import OpenLoopCLPass
from .BasePass import BasePass
from .sim.AddSimUtilFuncsPass import AddSimUtilFuncsPass
from .sim.DynamicSchedulePass import DynamicSchedulePass
from .sim.GenDAGPass import GenDAGPass
from .sim.SimpleSchedulePass import SimpleSchedulePass
from .sim.SimpleTickPass import SimpleTickPass
from .sim.WrapGreenletPass import WrapGreenletPass
from .tracing.CLLineTracePass import CLLineTracePass
from .tracing.CollectSignalPass import CollectSignalPass
from .tracing.LineTraceParamPass import LineTraceParamPass
from .tracing.PrintWavePass import PrintWavePass
from .tracing.VcdGenerationPass import VcdGenerationPass


# SimpleSim can be used when the UDG is a DAG
class SimpleSimPass( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    SimpleSchedulePass()( top )
    CLLineTracePass()( top )
    VcdGenerationPass()( top )
    CollectSignalPass()( top )
    PrintWavePass()( top )
    SimpleTickPass()( top )
    AddSimUtilFuncsPass()( top )
    LineTraceParamPass()( top )
    top.lock_in_simulation()

# This pass is created to be used for 2019 isca tutorial.
# Now we can always use this
class SimulationPass( BasePass ):
  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    CLLineTracePass()( top )
    DynamicSchedulePass()( top )
    VcdGenerationPass()( top )
    CollectSignalPass()( top )
    PrintWavePass()( top )
    SimpleTickPass()( top )
    AddSimUtilFuncsPass()( top )
    LineTraceParamPass()( top )
    top.lock_in_simulation()

class AutoTickSimPass( BasePass ):
  def __init__( s, print_line_trace=True ):
    s.print_line_trace = print_line_trace

  def __call__( s, top ):
    top.elaborate()
    GenDAGPass()( top )
    WrapGreenletPass()( top )
    OpenLoopCLPass( s.print_line_trace )( top )
    top.lock_in_simulation()



#=========================================================================
# RTLIRType.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 31, 2019
"""RTLIR instance types and generation methods.

This file contains the definitions of RTLIR instance types and methods
that generate RTLIR instances. Each instance of the non-abstract class
listed in this module is a type instance or simply a type in the RTLIR
type system. Signal types is parameterized by data types defined in the
RTLIR data type module.
"""
import copy
import inspect

import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits, is_bitstruct_inst

from ..errors import RTLIRConversionError
from ..util.utility import collect_objs
from . import RTLIRDataType as rdt


class BaseRTLIRType:
  """Base abstract class for all RTLIR instance types."""
  def __ne__( s, other ):
    return not s.__eq__( other )

class NoneType( BaseRTLIRType ):
  """Type for not yet typed RTLIR temporary variables."""
  def __eq__( s, other ):
    return isinstance( other, NoneType )

  def __hash__( s ):
    return hash(type(s))

  def __str__( s ):
    return 'NoneType'

  def __repr__( s ):
    return 'NoneType'

class Array( BaseRTLIRType ):
  """Unpacked RTLIR array type."""
  def __init__( s, dim_sizes, sub_type, obj = None, unpacked = False ):
    assert isinstance( sub_type, BaseRTLIRType ), \
      f"array subtype {sub_type} is not RTLIR type!"
    assert not isinstance( sub_type, Array ), \
      f"array subtype {sub_type} should not be array RTLIR type!"
    assert len( dim_sizes ) >= 1, "array should be non-empty!"
    assert sum( dim_sizes ) > 0, "array should have at least one element!"
    s.dim_sizes = dim_sizes
    s.sub_type = sub_type
    s.unpacked = unpacked
    s.obj = obj

  def _is_unpacked( s ):
    return s.unpacked

  def __eq__( s, other ):
    if not isinstance( other, Array ): return False
    if s.dim_sizes != other.dim_sizes: return False
    return s.sub_type == other.sub_type

  def __hash__( s ):
    return hash((type(s), tuple(s.dim_sizes), s.sub_type))

  def get_obj( s ):
    return s.obj

  def get_next_dim_type( s ):
    if len( s.dim_sizes ) == 1: return copy.copy( s.sub_type )
    _s = copy.copy( s )
    _s.dim_sizes = s.dim_sizes[1:]
    return _s

  def get_dim_sizes( s ):
    return s.dim_sizes

  def get_sub_type( s ):
    return s.sub_type

  def __call__( s, obj ):
    """Return if obj be cast into type `s`."""
    return s == obj

  def __str__( s ):
    return 'Array'

  def __repr__( s ):
    return f'Array{s.dim_sizes} of {s.sub_type}'

class Signal( BaseRTLIRType ):
  """Signal abstract RTLIR instance type.

  A Signal can be a Port, a Wire, or a Const.
  """
  def __init__( s, dtype, unpacked = False ):
    assert isinstance( dtype, rdt.BaseRTLIRDataType ), \
      f"signal parameterized by non-RTLIR data type {dtype}!"
    s.dtype = dtype
    s.unpacked = unpacked

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def is_packed_indexable( s ):
    return isinstance( s.dtype, rdt.PackedArray )

  def get_dtype( s ):
    return s.dtype

  def _is_unpacked( s ):
    return s.unpacked

class Port( Signal ):
  """Port RTLIR instance type."""
  def __init__( s, direction, dtype, unpacked = False ):
    super().__init__( dtype, unpacked )
    s.direction = direction

  def __eq__( s, other ):
    return isinstance(other, Port) and s.dtype == other.dtype and \
           s.direction == other.direction

  def __hash__( s ):
    return hash((type(s), s.dtype, s.direction))

  def __str__( s ):
    return 'Port'

  def __repr__( s ):
    return f'Port of {s.dtype}'

  def get_direction( s ):
    return s.direction

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable port!"
    return Port( s.direction, s.dtype.get_next_dim_type(), s.unpacked )

class Wire( Signal ):
  """Wire RTLIR instance type."""
  def __init__( s, dtype, unpacked = False ):
    super().__init__( dtype, unpacked )

  def __eq__( s, other ):
    return isinstance(other, Wire) and s.dtype == other.dtype

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def __str__( s ):
    return 'Wire'

  def __repr__( s ):
    return f'Wire of {s.dtype}'

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable wire!"
    return Wire( s.dtype.get_next_dim_type(), s.unpacked )

class NetWire( Signal ):
  """NetWire RTLIR instance type.

  Objects of this class are generated by the behavioral RTLIR type check pass
  to mark an expression as not available for any signal operations ( i.e.
  bit selection, part selection, attribute access ). This is useful because
  some backend tools ( like verilator ) does not support indexing/slicing over
  an already indexed/sliced signal ( the "net wire" ).
  """
  def __init__( s, dtype, unpacked = False ):
    super().__init__( dtype, unpacked )

  def __eq__( s, other ):
    return isinstance(other, NetWire) and s.dtype == other.dtype

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def __str__( s ):
    return 'NetWire'

  def __repr__( s ):
    return f'NetWire of {s.dtype}'

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable net wire!"
    return NetWire( s.dtype.get_next_dim_type(), s.unpacked )

class Const( Signal ):
  """Const RTLIR instance type."""
  def __init__( s, dtype, obj = None, unpacked = False ):
    super().__init__( dtype, unpacked )
    s.obj = obj

  def __eq__( s, other ):
    return isinstance(other, Const) and s.dtype == other.dtype

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def __str__( s ):
    return 'Const'

  def __repr__( s ):
    return f'Const of {s.dtype}'

  def get_object( s ):
    return s.obj

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable constant!"
    return Const( s.dtype.get_next_dim_type(), s.unpacked )

class InterfaceView( BaseRTLIRType ):
  """RTLIR instance type for a view of an interface."""
  def __init__( s, name, properties, obj = None, unpacked = False ):
    s.name = name
    s.properties = properties
    s.obj = obj
    s.cls = obj.__class__
    s.unpacked = unpacked

    if obj is not None:
      try:
        s.args = obj._dsl.args
        s.kwargs = obj._dsl.kwargs
      except AttributeError:
        assert False, f"interface object {s.obj} is not constructed!"

    # Sanity check
    for name, rtype in properties.items():
      assert isinstance(name, str) and _is_of_type(rtype, (Port, InterfaceView)), \
        f"invalid attribute {name} of interface {s.name}: only ports and interfaces allowed!"

  def __str__( s ):
    return f'InterfaceView {s.name}'

  def __repr__( s ):
    return f'InterfaceView {s.name}'

  def _is_unpacked( s ):
    return s.unpacked

  def __eq__( s, other ):
    return isinstance(other, InterfaceView) and s.name == other.name

  def __hash__( s ):
    return hash((type(s), s.name))

  def get_name( s ):
    return s.name

  def get_class( s ):
    return s.cls

  def get_args( s ):
    if s.obj is not None:
      return ( s.args, s.kwargs )
    else:
      return ( [], {} )

  def get_input_ports( s ):
    return sorted([id__port for id__port in s.properties.items() if id__port[1].direction == 'input'], key = lambda kv: kv[0])

  def get_output_ports( s ):
    return sorted([id__port1 for id__port1 in s.properties.items() if id__port1[1].direction == 'output'], key = lambda kv: kv[0])

  def has_property( s, p ):
    return p in s.properties

  def get_property( s, p ):
    return s.properties[ p ]

  def get_all_ports( s ):
    return sorted([name_port for name_port in s.properties.items() if isinstance( name_port[1], Port )], key = lambda kv: kv[0])

  def get_all_ports_packed( s ):
    return sorted([id__t2 for id__t2 in s.properties.items() if ( isinstance( id__t2[1], Port ) and not id__t2[1]._is_unpacked() ) or \
        ( isinstance( id__t2[1], Array ) and isinstance( id__t2[1].get_sub_type(), Port ) \
          and not id__t2[1]._is_unpacked() )], key = lambda kv: kv[0])

  def get_all_properties( s ):
    return sorted( s.properties.items(), lambda kv: kv[0] )

  def get_all_properties_packed( s ):
    return sorted( [x for x in s.properties.items() if not x[1]._is_unpacked()],
        key = lambda kv: kv[0] )

class Component( BaseRTLIRType ):
  """RTLIR instance type for a component."""
  def __init__( s, obj, properties, unpacked = False ):
    s.name = obj.__class__.__name__
    s.params = s._gen_parameters( obj )
    s.properties = properties
    s.unpacked = unpacked
    cls = obj.__class__
    try:
      file_name = inspect.getsourcefile( cls )
      s.file_info = f"{file_name}"
      # line_no = inspect.getsourcelines( cls )[1]
      # s.file_info = f"{file_name}:{line_no}"
    except OSError:
      s.file_info = f"Dynamically generated component {cls.__name__}"

  def _gen_parameters( s, obj ):
    # s.argspec: static code reflection results
    # _dsl.args: all unnamed arguments supplied to construct()
    # _dsl.kwargs: all named arguments supplied to construct()
    try:
      argspec = inspect.getfullargspec( obj.construct )
      assert not argspec.varkw, "keyword args are not allowed for construct!"
      assert not argspec.kwonlyargs, "keyword args are not allowed for construct!"
    except AttributeError:
      argspec = inspect.getargspec( obj.construct )
      assert not argspec.keywords, "keyword args are not allowed for construct!"
    assert not argspec.varargs, "varargs are not allowed for construct!"
    arg_names = argspec.args[1:]

    defaults = argspec.defaults or ()
    num_args = len(arg_names)
    num_supplied = len(obj._dsl.args) + len(obj._dsl.kwargs)
    num_defaults = len(defaults)

    # No default values: each arg is either keyword or unnamed
    # Has default values: num. supplied values + num. of defaults >= num. args
    assert num_args == num_supplied or num_args <= num_supplied + num_defaults, \
        "internal error: fail to parse the arguments!"
    use_defaults = num_args != num_supplied

    ret = []
    # Handle method construct arguments
    for idx, arg_name in enumerate(arg_names):

      # Use values from _dsl.args
      if idx < len(obj._dsl.args):
        ret.append((arg_name, obj._dsl.args[idx]))

      # Use values from _dsl.kwargs
      elif arg_name in obj._dsl.kwargs:
        ret.append((arg_name, obj._dsl.kwargs[arg_name]))

      # Use default values
      else:
        assert use_defaults, "internal error: didn't expect to use default values!"
        ret.append((arg_name, defaults[idx-len(arg_names)]))

    return ret

  def _is_unpacked( s ):
    return s.unpacked

  def __eq__( s, other ):
    return isinstance(other, Component) and s.name == other.name and \
           s.params == other.params

  def __hash__( s ):
    return hash((type(s), s.name, tuple(s.params)))

  def __str__( s ):
    return 'Component'

  def __repr__( s ):
    return f'Component {s.name}'

  def get_name( s ):
    return s.name

  def get_file_info( s ):
    return s.file_info

  def get_params( s ):
    return s.params

  def get_ports( s ):
    return sorted([id__port3 for id__port3 in s.properties.items() if isinstance( id__port3[1], Port )], key = lambda kv: kv[0])

  def get_ports_packed( s ):
    return sorted([id__t4 for id__t4 in s.properties.items() if ( isinstance( id__t4[1], Port ) and not id__t4[1]._is_unpacked() ) or \
        ( isinstance( id__t4[1], Array ) and isinstance( id__t4[1].get_sub_type(), Port ) \
          and not id__t4[1]._is_unpacked() )], key = lambda kv: kv[0])

  def get_wires( s ):
    return sorted([id__wire for id__wire in s.properties.items() if isinstance( id__wire[1], Wire )], key = lambda kv: kv[0])

  def get_wires_packed( s ):
    return sorted([id__t5 for id__t5 in s.properties.items() if ( isinstance( id__t5[1], Wire ) and not id__t5[1]._is_unpacked() ) or \
        ( isinstance( id__t5[1], Array ) and isinstance( id__t5[1].get_sub_type(), Wire ) \
          and not id__t5[1]._is_unpacked() )], key = lambda kv: kv[0])

  def get_consts( s ):
    return sorted([id__const for id__const in s.properties.items() if isinstance( id__const[1], Const )], key = lambda kv: kv[0])

  def get_consts_packed( s ):
    return sorted([id__t6 for id__t6 in s.properties.items() if ( isinstance( id__t6[1], Const ) and not id__t6[1]._is_unpacked() ) or \
        ( isinstance( id__t6[1], Array ) and isinstance( id__t6[1].get_sub_type(), Const ) \
          and not id__t6[1]._is_unpacked() )], key = lambda kv: kv[0])

  def get_ifc_views( s ):
    return sorted([id__ifc for id__ifc in s.properties.items() if isinstance( id__ifc[1], InterfaceView )], key = lambda kv: kv[0])

  def get_ifc_views_packed( s ):
    return sorted([id__t7 for id__t7 in s.properties.items() if ( isinstance( id__t7[1], InterfaceView ) and not id__t7[1]._is_unpacked() ) or \
        ( isinstance( id__t7[1], Array ) and isinstance( id__t7[1].get_sub_type(),InterfaceView ) \
          and not id__t7[1]._is_unpacked() )], key = lambda kv: kv[0])

  def get_subcomps( s ):
    return sorted([id__subcomp for id__subcomp in s.properties.items() if isinstance( id__subcomp[1], Component )], key = lambda kv: kv[0])

  def get_subcomps_packed( s ):
    return sorted([id__t9 for id__t9 in s.properties.items() if ( isinstance( id__t9[1], Component ) and not id__t9[1]._is_unpacked() ) or \
        ( isinstance( id__t9[1], Array ) and isinstance( id__t9[1].get_sub_type(), Component ) \
          and not id__t9[1]._is_unpacked() )], key = lambda kv: kv[0])

  def has_property( s, p ):
    return p in s.properties

  def get_property( s, p ):
    return s.properties[ p ]

  def get_all_properties( s ):
    return s.properties

#-------------------------------------------------------------------------
# Internal Methods
#-------------------------------------------------------------------------

def _is_rtlir_ifc_convertible( obj ):
  valid_ifc_attributes = ( dsl.InPort, dsl.OutPort, dsl.Interface )
  if isinstance( obj, list ):
    while isinstance( obj, list ):
      # Empty lists will be dropped
      if len( obj ) == 0:
        return True
      # assert len( obj ) > 0, f"one dimension of {obj} is 0!"
      obj = obj[0]
    return _is_rtlir_ifc_convertible( obj )
  elif isinstance( obj, valid_ifc_attributes ):
    return True
  else:
    return False

def _freeze( obj ):
  if isinstance( obj, list ):
    return tuple( _freeze(value) for value in obj )
  else:
    return obj

def _unpack( id_, Type ):
  if not isinstance( Type, Array ): return [ ( id_, Type ) ]
  ret = []
  for idx in range( Type.get_dim_sizes()[0] ):
    ret.append( ( f'{id_}[{idx}]', Type.get_next_dim_type() ) )
    ret.extend(_unpack(f'{id_}[{idx}]', Type.get_next_dim_type()))
  return ret

def _add_packed_instances( id_, Type, properties ):
  assert isinstance( Type, Array ), f"{Type} is not an array type!"
  for _id, _Type in _unpack( id_, Type ):
    assert hasattr( _Type, 'unpacked' ), f"{_Type} {_id} is not unpacked!"
    _Type.unpacked = True
    properties[ _id ] = _Type

def _handle_Array( _id, _obj ):
  obj = _obj
  if len( obj ) == 0:
    return None
  ref_type = get_rtlir( obj[0] )
  assert all( get_rtlir(i) == ref_type for i in obj ), \
    f'all elements of array {obj} must have the same type {ref_type}!'
  dim_sizes = []
  while isinstance( obj, list ):
    if len( obj ) == 0:
      return None
    dim_sizes.append( len( obj ) )
    obj = obj[0]
  if isinstance( obj, ( int, Bits ) ):
    return Array( dim_sizes, get_rtlir( obj ), _obj )
  else:
    return Array( dim_sizes, get_rtlir( obj ) )

def _handle_InPort( p_id, obj ):
  return Port( 'input', rdt.get_rtlir_dtype( obj ) )

def _handle_OutPort( p_id, obj ):
  return Port( 'output', rdt.get_rtlir_dtype( obj ) )

def _handle_Wire( w_id, obj ):
  return Wire( rdt.get_rtlir_dtype( obj ) )

def _handle_Const( c_id, obj ):
  return Const( rdt.get_rtlir_dtype( obj ), obj )

def _handle_Interface( i_id, obj ):
  properties = {}
  collected_objs = collect_objs( obj, object )
  for _id, _obj in collected_objs:
    if _is_rtlir_ifc_convertible( _obj ):
      _obj_type = get_rtlir( _obj )
      if _obj_type is not None:
        properties[ _id ] = _obj_type
        if isinstance( _obj_type, Array ):
          _add_packed_instances( _id, _obj_type, properties )
    # TODO: Figure out a way to inform user of dropped attributes without
    # flooding STDOUT
    # else:
      # err_msg = \
# """\
 # - Note: {} attribute {} of {} was dropped during conversion to RTLIR because it is
         # not an interface, a port, or a list of them. \
# """
      # print( err_msg.format( _id, _obj, i_id ) )
  return InterfaceView( obj.__class__.__name__, properties, obj )

def _handle_Component( c_id, obj ):
  properties = {}
  collected_objs = collect_objs( obj, object )
  for _id, _obj in collected_objs:
    # Untranslatable attributes will be ignored
    if is_rtlir_convertible( _obj ):
      _obj_type = get_rtlir( _obj )
      if _obj_type is not None:
        properties[ _id ] = _obj_type
        if isinstance( _obj_type, Array ):
          _add_packed_instances( _id, _obj_type, properties )
    # TODO: Figure out a way to inform user of dropped attributes without
    # flooding STDOUT
    # else:
      # err_msg = \
# """\
 # - Note: {} attribute {} of {} was dropped during conversion to RTLIR because it is
         # not a port, a, wire, an interface, a component, a constantor, or a
         # list of them. \
# """
      # print( err_msg.format( _id, _obj, c_id ) )
  return Component( obj, properties )

def _is_of_type( obj, Type ):
  """Return True is `obj` is of RTLIR type `Type`."""
  if isinstance( obj, Type ):
    return True
  if isinstance( obj, Array ) and isinstance( obj.get_sub_type(), Type ):
    return True
  return False

_RTLIR_ifc_handlers = [
  ( list,          _handle_Array ),
  ( dsl.InPort,    _handle_InPort ),
  ( dsl.OutPort,   _handle_OutPort ),
  ( dsl.Interface, _handle_Interface ),
]

_RTLIR_handlers = [
  ( list,          _handle_Array ),
  ( dsl.InPort,    _handle_InPort ),
  ( dsl.OutPort,   _handle_OutPort ),
  ( dsl.Wire,      _handle_Wire ),
  ( ( int, Bits ), _handle_Const ),
  ( dsl.Interface, _handle_Interface ),
  ( dsl.Component, _handle_Component ),
]

#-------------------------------------------------------------------------
# Public APIs
#-------------------------------------------------------------------------

def get_component_ifc_rtlir( obj ):
  """Return the RTLIR of the interfaces of component `obj`."""
  primitive_types = tuple( x[0] for x in _RTLIR_ifc_handlers[1:] )

  def _is_interface( id_, obj ):
    _type = type(obj)
    if isinstance( obj, primitive_types ):
      return True
    if not isinstance( obj, list ):
      return False
    while isinstance( obj, list ):
      assert len( obj ) > 0, f"{id_} is an empty list!"
      obj = obj[0]
    return isinstance( obj, primitive_types )

  def _get_ifc_rtlir( id_, obj ):
    for Type, handler in _RTLIR_ifc_handlers:
      if isinstance( obj, Type ):
        return handler( id_, obj )
    return None

  try:
    assert isinstance(obj, dsl.Component), \
      "the given object is not a PyMTL component!"
    properties = {}
    collected_objs = collect_objs( obj, object )
    for _id, _obj in collected_objs:
      if _is_interface( _id, _obj ):
        _obj_type = _get_ifc_rtlir( _id, _obj )
        if _obj_type is not None:
          properties[ _id ] = _obj_type
          if isinstance( _obj_type, Array ):
            _add_packed_instances( _id, _obj_type, properties )
    return Component( obj, properties )
  except AssertionError as e:
    msg = '' if e.args[0] is None else e.args[0]
    raise RTLIRConversionError( obj, msg )

def is_rtlir_convertible( obj ):
  """Return if `obj` can be converted into an RTLIR instance."""
  pymtl_constructs = (
    dsl.InPort, dsl.OutPort, dsl.Wire,
    Bits, dsl.Interface, dsl.Component,
  )
  # TODO: improve this long list of isinstance check
  if isinstance( obj, list ):
    while isinstance( obj, list ):
      # Empty lists will be dropped
      if len( obj ) == 0:
        return True
      obj = obj[0]
    return is_rtlir_convertible( obj )
  elif isinstance( obj, pymtl_constructs ):
    return True
  elif is_bitstruct_inst(obj):
    return True
  elif isinstance( obj, int ):
    return True
  else:
    return False

_rtlir_cache = {}

def get_rtlir( _obj ):
  """Return an RTLIR instance corresponding to `obj`."""
  obj = _freeze( _obj )
  if obj in _rtlir_cache:
    return _rtlir_cache[ obj ]
  else:
    try:
      for Type, handler in _RTLIR_handlers:
        if isinstance( _obj, Type ):
          ret = _rtlir_cache[ obj ] = handler( "<name not available>", _obj )
          return ret
      if is_bitstruct_inst( _obj ):
        ret = _rtlir_cache[ obj ] = _handle_Const( "<name not available>", _obj )
        return ret

      # Cannot convert `obj` into RTLIR representation
      assert False, f'unrecognized object {_obj}!'
    except AssertionError as e:
      msg = '' if e.args[0] is None else e.args[0]
      raise RTLIRConversionError( _obj, msg )



"""
========================================================================
SendRecvIfc.py
========================================================================
RTL implementation of en/rdy micro-protocol.

Author: Yanghui Ou, Shunning Jiang
  Date: May 5, 2019
"""
from copy import deepcopy

import greenlet

from pymtl3 import *
from pymtl3.dsl.errors import InvalidConnectionError
from pymtl3.stdlib.connects import connect_pairs

#-------------------------------------------------------------------------
# RecvIfcRTL
#-------------------------------------------------------------------------

class RecvIfcRTL( CalleeIfcRTL ):

  def construct( s, Type ):
    super().construct( en=True, rdy=True, MsgType=Type, RetType=None )

  def connect( s, other, parent ):

    # We are doing SendCL (other) -> [ RecvCL -> SendRTL ] -> RecvRTL (s)
    # SendCL is a caller interface
    if isinstance( other, CallerIfcCL ):
      m = RecvCL2SendRTL( s.MsgType )

      if hasattr( parent, "RecvCL2SendRTL_count" ):
        count = parent.RecvCL2SendRTL_count
        setattr( parent, "RecvCL2SendRTL_" + str( count ), m )
      else:
        parent.RecvCL2SendRTL_count = 0
        parent.RecvCL2SendRTL_0 = m

      connect_pairs(
        other,  m.recv,
        m.send.msg, s.msg,
        m.send.en,  s.en,
        m.send.rdy, s.rdy
      )
      parent.RecvCL2SendRTL_count += 1
      return True

    elif isinstance( other, CalleeIfcCL ):
      if s._dsl.level <= other._dsl.level:
        raise InvalidConnectionError(
            "CL2RTL connection is not supported between RecvIfcRTL"
            " and CalleeIfcCL.\n"
            "          - level {}: {} (class {})\n"
            "          - level {}: {} (class {})".format(
                s._dsl.level, repr( s ), type( s ), other._dsl.level,
                repr( other ), type( other ) ) )

      m = RecvCL2SendRTL( s.MsgType )

      if hasattr( parent, "RecvCL2SendRTL_count" ):
        count = parent.RecvCL2SendRTL_count
        setattr( parent, "RecvCL2SendRTL_" + str( count ), m )
      else:
        parent.RecvCL2SendRTL_count = 0
        parent.RecvCL2SendRTL_0 = m

      connect_pairs(
        other,  m.recv,
        m.send.msg, s.msg,
        m.send.en,  s.en,
        m.send.rdy, s.rdy
      )
      parent.RecvCL2SendRTL_count += 1
      return True

    return False

#-------------------------------------------------------------------------
# SendIfcRTL
#-------------------------------------------------------------------------

class SendIfcRTL( CallerIfcRTL ):

  def construct( s, Type ):
    super().construct( en=True, rdy=True, MsgType=Type, RetType=None )

  def connect( s, other, parent ):

    # We are doing SendRTL (s) -> [ RecvRTL -> SendCL ] -> RecvCL (other)
    # RecvCL is a callee interface
    if isinstance( other, CalleeIfcCL ):
      m = RecvRTL2SendCL( s.MsgType )

      if hasattr( parent, "RecvRTL2SendCL_count" ):
        count = parent.RecvRTL2SendCL_count
        setattr( parent, "RecvRTL2SendCL_" + str( count ), m )
      else:
        parent.RecvRTL2SendCL_count = 0
        parent.RecvRTL2SendCL_0 = m

      connect_pairs(
        m.send, other,
        s.msg, m.recv.msg,
        s.en,  m.recv.en,
        s.rdy, m.recv.rdy,
      )
      parent.RecvRTL2SendCL_count += 1
      return True

    return False


class SendIfcFL( CallerIfcFL ):
  pass

  def connect( s, other, parent ):

    # We are doing SendFL (s) -> [ RecvFL -> SendCL ] -> RecvCL (s)
    # SendCL is a caller interface
    # FIXME direction
    if isinstance( other, CallerIfcCL ) or \
       isinstance( other, CalleeIfcCL ):
      m = RecvFL2SendCL()

      if hasattr( parent, "RecvFL2SendCL_count" ):
        count = parent.RecvFL2SendCL_count
        setattr( parent, "RecvFL2SendCL_" + str( count ), m )
      else:
        parent.RecvFL2SendCL_count = 0
        parent.RecvFL2SendCL_0 = m

      connect_pairs(
        s,      m.recv,
        m.send, other,
      )
      parent.RecvFL2SendCL_count += 1
      return True

    return False

class RecvIfcFL( CalleeIfcFL ):
  pass

"""
========================================================================
Send/RecvIfc adapters
========================================================================
CL/RTL adapters for send/recv interface.

Author : Yanghui Ou
  Date : Mar 07, 2019
"""

#-------------------------------------------------------------------------
# RecvCL2SendRTL
#-------------------------------------------------------------------------

class RecvCL2SendRTL( Component ):

  def construct( s, MsgType ):

    # Interface

    s.send = SendIfcRTL( MsgType )

    s.entry = None

    @s.update
    def up_clear():
      if s.send.en: # constraints reverse this
        s.entry = None

    @s.update
    def up_send_rtl():
      if s.entry is None:
        s.send.en  = b1( 0 )
        s.send.msg = MsgType()
      else:
        s.send.en  = b1( s.send.rdy )
        s.send.msg = s.entry

    s.add_constraints(
      U( up_clear )   < WR( s.send.en ),
      U( up_clear )   < M( s.recv ),
      U( up_clear )   < M( s.recv.rdy ),
      M( s.recv )     < U( up_send_rtl ),
      M( s.recv.rdy ) < U( up_send_rtl )
    )

  @non_blocking( lambda s : s.entry is None )
  def recv( s, msg ):
    s.entry = deepcopy(msg)

  def line_trace( s ):
    return "{}(){}".format( s.recv, s.send )

#-------------------------------------------------------------------------
# RecvRTL2SendCL
#-------------------------------------------------------------------------

class RecvRTL2SendCL( Component ):

  def construct( s, MsgType ):

    # Interface

    s.recv = RecvIfcRTL( MsgType )
    s.send = CallerIfcCL()

    s.sent_msg = None
    s.send_rdy = False

    @s.update
    def up_recv_rtl_rdy():
      s.send_rdy = s.send.rdy() and not s.reset
      s.recv.rdy = b1( 1 ) if s.send.rdy() and not s.reset else b1( 0 )

    @s.update
    def up_send_cl():
      s.sent_msg = None
      if s.recv.en:
        s.send( s.recv.msg )
        s.sent_msg = s.recv.msg

    s.add_constraints( U( up_recv_rtl_rdy ) < U( up_send_cl ) )

  def line_trace( s ):
    return "{}(){}".format(
      s.recv.line_trace(),
      enrdy_to_str( s.sent_msg, s.sent_msg is not None, s.send_rdy )
    )

#-------------------------------------------------------------------------
# RecvFL2SendCL
#-------------------------------------------------------------------------

class RecvFL2SendCL( Component ):

  @blocking
  def recv( s, msg ):
    while not s.send.rdy():
      greenlet.getcurrent().parent.switch(0)
    assert s.send.rdy()
    s.send( msg )

  def construct( s ):

    # Interface

    s.recv = RecvIfcFL( method=s.recv )
    s.send = CallerIfcCL()

    s.add_constraints( M( s.recv ) == M( s.send ) )

  def line_trace( s ):
    return "{}(){}".format( s.recv, s.send )



"""
These objects are only for developers to use. Other objects are exposed
to users in pymtl/__init__.py
"""
from .Component import Component
from .ComponentLevel3 import connect
from .ComponentLevel5 import method_port
from .ComponentLevel6 import non_blocking
from .ComponentLevel7 import blocking
from .Connectable import (
    BlockingIfc,
    CalleeIfcCL,
    CalleeIfcFL,
    CalleeIfcRTL,
    CalleePort,
    CallerIfcCL,
    CallerIfcFL,
    CallerIfcRTL,
    CallerPort,
    CallIfcRTL,
    Const,
    InPort,
    Interface,
    MethodPort,
    NonBlockingIfc,
    OutPort,
    Signal,
    Wire,
)
from .ConstraintTypes import RD, WR, M, U
from .Placeholder import Placeholder



"""
========================================================================
Tests for CL queues
========================================================================

Author: Yanghui Ou
  Date: Mar 20, 2019
"""
import pytest

from pymtl3 import *
from pymtl3.stdlib.test.test_sinks import TestSinkCL
from pymtl3.stdlib.test.test_srcs import TestSrcCL

from .queues import BypassQueueCL, NormalQueueCL, PipeQueueCL

#-------------------------------------------------------------------------
# TestHarness
#-------------------------------------------------------------------------

class TestHarness( Component ):

  def construct( s, MsgType, DutType, src_msgs, sink_msgs ):

    s.src     = TestSrcCL ( MsgType, src_msgs )
    s.dut     = DutType()
    s.sink    = TestSinkCL( MsgType, sink_msgs )

    connect( s.src.send, s.dut.enq )

    @s.update
    def up_deq_send():
      if s.dut.deq.rdy() and s.sink.recv.rdy():
        s.sink.recv( s.dut.deq() )

  def done( s ):
    return s.src.done() and s.sink.done()

  def line_trace( s ):
    return "{} ({}) {}".format(
      s.src.line_trace(), s.dut.line_trace(), s.sink.line_trace() )

#-------------------------------------------------------------------------
# run_sim
#-------------------------------------------------------------------------

def run_sim( th, max_cycles=100 ):

  # Create a simulator

  th.apply( SimulationPass() )

  # Run simulation

  print("")
  ncycles = 0
  print("{:2}:{}".format( ncycles, th.line_trace() ))
  while not th.done() and ncycles < max_cycles:
    th.tick()
    ncycles += 1
    print("{:2}:{}".format( ncycles, th.line_trace() ))

  # Check timeout

  assert ncycles < max_cycles

  th.tick()
  th.tick()
  th.tick()

#-------------------------------------------------------------------------
# Test cases
#-------------------------------------------------------------------------

test_msgs = [ Bits16( 0 ), Bits16( 1 ), Bits16( 2 ), Bits16( 3 ) ]

arrival_pipe   = [ 2, 3, 4, 5 ]
arrival_bypass = [ 1, 2, 3, 4 ]
arrival_normal = [ 2, 4, 6, 8 ]

def test_pipe_simple():
  th = TestHarness( Bits16, PipeQueueCL, test_msgs, test_msgs )
  th.set_param( "top.dut.construct", num_entries=1 )
  th.set_param( "top.sink.construct", arrival_time=arrival_pipe )
  run_sim( th )

def test_bypass_simple():
  th = TestHarness( Bits16, BypassQueueCL, test_msgs, test_msgs )
  th.set_param( "top.dut.construct", num_entries=1 )
  th.set_param( "top.sink.construct", arrival_time=arrival_bypass )
  run_sim( th )

def test_normal_simple():
  th = TestHarness( Bits16, NormalQueueCL, test_msgs, test_msgs )
  th.set_param( "top.dut.construct", num_entries=1 )
  th.set_param( "top.sink.construct", arrival_time=arrival_normal )
  run_sim( th )

def test_normal2_simple():
  th = TestHarness( Bits16, NormalQueueCL, test_msgs, test_msgs )
  th.set_param( "top.dut.construct", num_entries=2 )
  th.set_param( "top.sink.construct", arrival_time=arrival_pipe )
  run_sim( th )

@pytest.mark.parametrize(
  ( 'QType', 'qsize', 'src_init', 'src_intv',
    'sink_init', 'sink_intv', 'arrival_time' ),
  [
    ( PipeQueueCL,   2, 1, 1, 0, 0, [ 3, 5,  7,  9 ] ),
    ( BypassQueueCL, 1, 0, 4, 3, 1, [ 3, 6, 11, 16 ] ),
    ( NormalQueueCL, 1, 0, 0, 5, 0, [ 5, 7,  9, 11 ] )
  ]
)
def test_delay( QType, qsize, src_init, src_intv,
                sink_init, sink_intv, arrival_time ):
  th = TestHarness( Bits16, QType, test_msgs, test_msgs )
  th.set_param( "top.src.construct",
    initial_delay  = src_init,
    interval_delay = src_intv,
  )
  th.set_param( "top.dut.construct", num_entries=qsize )
  th.set_param( "top.sink.construct",
    initial_delay  = sink_init,
    interval_delay = sink_intv,
    arrival_time   = arrival_time,
  )
  run_sim( th )



"""
========================================================================
ComponentLevel3_test.py
========================================================================

Author : Shunning Jiang
Date   : Dec 25, 2017
"""
from collections import deque

from pymtl3.datatypes import Bits8, Bits10, Bits32, bitstruct
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3, connect
from pymtl3.dsl.Connectable import InPort, OutPort, Wire
from pymtl3.dsl.ConstraintTypes import WR, U
from pymtl3.dsl.errors import (
    InvalidConnectionError,
    MultiWriterError,
    PyMTLDeprecationError,
    UpblkFuncSameNameError,
)

from .sim_utils import simple_sim_pass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  T, time = 0, 20
  while not A.done() and T < time:
    A.tick()
    print(A.line_trace())
    T += 1

MUX_SEL_0 = 0
MUX_SEL_1 = 1

class TestSource( ComponentLevel3 ):

  def construct( s, Type, input_ ):
    assert type(input_) == list, "TestSrc only accepts a list of inputs!"
    s.input_ = deque([ Type(x) for x in input_ ]) # deque.popleft() is faster

    s.out = OutPort(Type)

    @s.update
    def up_src():
      if not s.input_:
        s.out = Type()
      else:
        s.out = s.input_.popleft()

  def done( s ):
    return not s.input_

  def line_trace( s ):
    return "%s" % s.out

class TestSink( ComponentLevel3 ):

  def construct( s, Type, answer ):
    assert type(answer) == list, "TestSink only accepts a list of outputs!"

    s.answer = deque( [ x if x == "*" else Type(x) for x in answer ] )
    s.in_ = InPort(Type)

    @s.update
    def up_sink():
      if not s.answer:
        assert False, "Simulation has ended"
      else:
        ref = s.answer.popleft()
        ans = s.in_

        assert ref == ans or ref == "*", "Expect {}, get {} instead".format(ref, ans)

  def done( s ):
    return not s.answer

  def line_trace( s ):
    return "%s" % s.in_

class Mux( ComponentLevel3 ):

  def construct( s, Type, ninputs ):
    s.in_ = [ InPort( Type ) for _ in range(ninputs) ]
    s.sel = InPort( int if Type is int else mk_bits( clog2(ninputs) ) )
    s.out = OutPort( Type )

    @s.update
    def up_mux():
      s.out = s.in_[ s.sel ]

  def line_trace( s ):  pass

def test_connect_list_const_idx():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux = Mux(int, 2)

      connect( s.mux.in_[MUX_SEL_0], s.src_in0.out )
      connect( s.mux.in_[MUX_SEL_1], s.src_in1.out )
      connect( s.mux.sel           , s.src_sel.out )
      connect( s.sink.in_          , s.mux.out     )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_list_const_idx_ifloordiv_sugar():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux = Mux(int, 2)

      s.mux.in_[MUX_SEL_0] //= s.src_in0.out
      s.mux.in_[MUX_SEL_1] //= s.src_in1.out
      s.mux.sel            //= s.src_sel.out
      s.sink.in_           //= s.mux.out

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )
def test_connect_list_idx_call():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux = Mux(int, 2)(
        out = s.sink.in_,
        in_ = { MUX_SEL_0: s.src_in0.out, MUX_SEL_1: s.src_in1.out },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_deep():

  class MuxWrap(ComponentLevel3):

    def construct( s ):
      s.in_ = [ InPort(int) for _ in range(2) ]
      s.sel = InPort(int)
      s.out = OutPort(int)

      s.mux = Mux(int, 2)(
        out = s.out,
        in_ = { 0: s.in_[0], 1: s.in_[1] },
        sel = s.sel,
      )

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux_wrap = MuxWrap()(
        out = s.sink.in_,
        in_ = { 0: s.src_in0.out, 1: s.src_in1.out },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_deep_connect():

  class MuxWrap3(ComponentLevel3):

    def construct( s ):
      s.in_ = [ InPort(int) for _ in range(2) ]
      s.sel = InPort(int)
      s.out = OutPort(int)

      s.mux1 = Mux(int, 2)(
        in_ = { 0: s.in_[0], 1: s.in_[1] },
        sel = s.sel,
      )
      s.mux2 = Mux(int, 2)(
        in_ = { 0: s.in_[0], 1: s.in_[1] },
        sel = s.sel,
      )
      s.mux3 = Mux(int, 2)(
        out = s.out,
        in_ = { 0: s.mux1.out, 1: s.mux2.out },
        sel = s.sel,
      )

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux_wrap = MuxWrap3()(
        out = s.sink.in_,
        in_ = { 0: s.src_in0.out, 1: s.src_in1.out },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_2d_array_vars_connect_impl():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( Bits32, [2,1,0,2,1,0] )
      s.sink = TestSink  ( Bits32, ["*",(5+6),(3+4),(1+2),
                                 (5+6),(3+4),(1+2)] )

      s.wire = [ [ Wire(Bits32) for _ in range(2)] for _ in range(2) ]
      connect( s.wire[0][0], s.src.out )

      @s.update
      def up_from_src():
        s.wire[0][1] = s.src.out + 1

      s.reg = Wire(Bits32)
      connect( s.wire[1][0], s.reg )

      @s.update_ff
      def up_reg():
        s.reg <<= s.wire[0][0] + s.wire[0][1]

      @s.update
      def upA():
        s.wire[1][1] = s.reg + 1

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire[1][0] + s.wire[1][1]

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             str(s.wire)+" r0=%s" % s.reg + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_lots_of_fan_connect():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( Bits32, [4,3,2,1,4,3,2,1] )
      s.sink = TestSink  ( Bits32, ["*",(5+5+6+6),(4+4+5+5),(3+3+4+4),(2+2+3+3),
                                     (5+5+6+6),(4+4+5+5),(3+3+4+4),(2+2+3+3)] )

      s.wire0 = Wire(Bits32)

      @s.update
      def up_from_src():
        s.wire0 = s.src.out + 1

      s.reg = Wire(Bits32)

      @s.update_ff
      def up_reg():
        s.reg <<= s.wire0

      s.wire1 = Wire(Bits32)
      s.wire2 = Wire(Bits32)

      connect( s.wire1, s.reg )

      @s.update
      def upA():
        s.wire2 = s.reg + 1

      s.wire3 = Wire(Bits32)
      s.wire4 = Wire(Bits32)

      connect( s.wire3, s.wire1 )
      connect( s.wire4, s.wire1 )

      s.wire5 = Wire(Bits32)
      s.wire6 = Wire(Bits32)

      connect( s.wire5, s.wire2 )
      connect( s.wire6, s.wire2 )

      s.wire7 = Wire(Bits32)
      s.wire8 = Wire(Bits32)

      @s.update
      def upD():
        s.wire7 = s.wire3 + s.wire6
        s.wire8 = s.wire4 + s.wire5

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire7 + s.wire8

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
            "w0={} > r0={}".format(s.wire0,s.reg) + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_plain():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( int, [4,3,2,1,4,3,2,1] )
      s.sink = TestSink  ( int, [5,4,3,2,5,4,3,2] )

      s.wire0 = Wire(int)

      @s.update
      def up_from_src():
        s.wire0 = s.src.out + 1

      connect( s.sink.in_, s.wire0 )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
            "w0=%s" % (s.wire0) + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_2d_array_vars_connect():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( int, [2,1,0,2,1,0] )
      s.sink = TestSink  ( int, ["*",(5+6),(3+4),(1+2),
                                     (5+6),(3+4),(1+2)] )

      s.wire = [ [ Wire(int) for _ in range(2)] for _ in range(2) ]
      connect( s.wire[0][0], s.src.out )

      @s.update
      def up_from_src():
        s.wire[0][1] = s.src.out + 1

      s.reg = Wire(int)
      connect( s.wire[1][0], s.reg )

      @s.update
      def up_reg():
        s.reg = s.wire[0][0] + s.wire[0][1]

      for i in range(2):
        s.add_constraints(
          U(up_reg) < WR(s.wire[0][i]), # up_reg reads  s.wire[0][i]
        )

      @s.update
      def upA():
        s.wire[1][1] = s.reg + 1

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire[1][0] + s.wire[1][1]

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             str(s.wire)+" r0=%s" % s.reg + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_const_same_level():

  class Top(ComponentLevel3):

    def construct( s ):

      s.a = Wire(int)
      connect( s.a, 0 )

      @s.update
      def up_printa():
        print(s.a)

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  _test_model( Top )

def test_connect_const_two_writer():

  class Top(ComponentLevel3):

    def construct( s ):

      s.a = Wire(int)
      connect( s.a, 0 )

      @s.update
      def up_printa():
        print(s.a)

      @s.update
      def up_writea():
        s.a = 123

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

def test_connect_list_idx_call():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [12,3,12,1] )

      s.mux = Mux(int, 2)(
        out = s.sink.in_,
        in_ = { MUX_SEL_0: s.src_in0.out },
        sel = s.src_sel.out,
      )
      connect( s.mux.in_[MUX_SEL_1], 12 )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_list_idx_const_in_call():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [12,3,12,1] )

      s.mux = Mux(int, 2)(
        out = s.sink.in_,
        in_ = { MUX_SEL_0: s.src_in0.out, MUX_SEL_1: 12 },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_top_level_inport():

  class Top( ComponentLevel3 ):

    def construct( s ):

      s.a = InPort(Bits10)
      s.b = Wire(Bits32)
      connect( s.a, s.b[0:10] )

      @s.update
      def up():
        print(s.b[10:32])

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  _test_model( Top )

def test_top_level_outport():

  class Top( ComponentLevel3 ):

    def construct( s ):

      s.a = OutPort(Bits10)
      s.b = Wire(Bits32)
      connect( s.a, s.b[9:19] )

      @s.update
      def up():
        s.b[0:10] = 1023

    def done( s ):
      return False

    def line_trace( s ):
      return str(int(s.a))

  A = Top()
  A.elaborate()
  simple_sim_pass( A )

  A.tick()
  trace = A.line_trace()
  print(" >>>", trace)
  assert trace == "1"

# This is a simplified test case from Peitian

def test_multiple_slices_are_net_writers():

  class A( ComponentLevel3 ):

    def construct( s ):
      s.in_  = InPort( Bits32 )
      s.out1 = OutPort( Bits8 )
      s.out2 = OutPort( Bits8 )

      connect( s.in_[0:8], s.out1[0:8] )
      connect( s.in_[0:4], s.out2[0:4] )

  a = A()
  a.elaborate()

# A similar case for struct

def test_multiple_fields_are_net_writers():

  @bitstruct
  class SomeMsg1:
    a: Bits8
    b: Bits32

  @bitstruct
  class SomeMsg2:
    c: Bits8

  class A( ComponentLevel3 ):

    def construct( s ):
      s.in_  = InPort( SomeMsg1 )
      s.out1 = OutPort( SomeMsg2 )
      s.out2 = OutPort( SomeMsg2 )

      connect( s.in_.a, s.out1.c )
      connect( s.in_.b[0:8], s.out2.c )

  a = A()
  a.elaborate()
  simple_sim_pass( a )

  a.in_ = SomeMsg1(4, 5)
  a.tick()
  assert a.out1.c == 4

# The counterpart of the above test

def test_multiple_fields_are_assigned():

  @bitstruct
  class SomeMsg1:
    a: Bits8
    b: Bits32

  @bitstruct
  class SomeMsg2:
    c: Bits8

  class A( ComponentLevel3 ):

    def construct( s ):
      s.in_  = InPort( SomeMsg1 )
      s.out1 = OutPort( SomeMsg2 )
      s.out2 = OutPort( SomeMsg2 )

      # connect( s.in_.a, s.out1.c )
      # connect( s.in_.b[0:8], s.out2.c )
      @s.update
      def up_pass():
        s.out1.c = s.in_.a
        s.out2.c = s.in_.b[0:8]

  a = A()
  a.elaborate()
  simple_sim_pass( a )

  a.in_ = SomeMsg1(4, 5)
  a.tick()
  assert a.out1.c == 4

def test_const_connect_struct_signal_to_int():

  class SomeMsg1:
    def __init__( s, a=0, b=0 ):
      s.a = Bits8(a)
      s.b = Bits32(b)

    def __eq__( s, other ):
      return s.a == other.a and s.b == other.b

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(SomeMsg1)
      connect( s.wire, 1 )

  try:
    x = Top()
    x.elaborate()
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

def test_const_connect_struct_signal_to_Bits():

  class SomeMsg1:
    def __init__( s, a=0, b=0 ):
      s.a = Bits8(a)
      s.b = Bits32(b)

    def __eq__( s, other ):
      return s.a == other.a and s.b == other.b

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(SomeMsg1)
      connect( s.wire, Bits32(1) )

  try:
    x = Top()
    x.elaborate()
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

def test_const_connect_Bits_signal_to_int():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(Bits32)
      connect( s.wire, 1 )

  x = Top()
  x.elaborate()
  print(x._dsl.consts)
  assert len(x._dsl.consts) == 1

  simple_sim_pass(x)
  x.tick()

def test_const_connect_int_signal_to_int():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(int)
      connect( s.wire, 1 )

  x = Top()
  x.elaborate()
  print(x._dsl.consts)
  assert len(x._dsl.consts) == 1

  simple_sim_pass(x)
  x.tick()

def test_const_connect_Bits_signal_to_Bits():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(Bits32)
      connect( s.wire, Bits32(0) )

  x = Top()
  x.elaborate()
  print(x._dsl.consts)
  assert len(x._dsl.consts) == 1

  simple_sim_pass(x)
  x.tick()

def test_const_connect_Bits_signal_to_mismatch_Bits():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(Bits32)
      connect( s.wire, Bits8(8) )

  try:
    x = Top()
    x.elaborate()
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

def test_const_connect_struct_signal_to_struct():

  @bitstruct
  class SomeMsg1:
    a: Bits8
    b: Bits32

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(SomeMsg1)
      connect( s.out, SomeMsg1(1,3) )

  x = Top()
  x.elaborate()
  simple_sim_pass(x)
  x.tick()
  assert x.out == SomeMsg1(1,3)

def test_const_connect_nested_struct_signal_to_struct():

  @bitstruct
  class SomeMsg1:
    a: Bits8
    b: Bits32

  @bitstruct
  class SomeMsg2:
    a: SomeMsg1
    b: Bits32

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(SomeMsg2)
      connect( s.out, SomeMsg2(SomeMsg1(1,2),3) )

  x = Top()
  x.elaborate()
  simple_sim_pass(x)
  x.tick()
  assert x.out == SomeMsg2(SomeMsg1(1,2),3)

def test_const_connect_cannot_handle_same_name_nested_struct():

  class A:
    @bitstruct
    class SomeMsg1:
      a: Bits8
      b: Bits32

  class B:
    @bitstruct
    class SomeMsg1:
      c: Bits8
      d: Bits32

  @bitstruct
  class SomeMsg2:
    a: A.SomeMsg1
    b: B.SomeMsg1

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(SomeMsg2)
      connect( s.out, SomeMsg2(A.SomeMsg1(1,2),B.SomeMsg1(3,4)) )

  x = Top()
  x.elaborate()
  try:
    simple_sim_pass(x)
  except AssertionError as e:
    print(e)
    assert str(e) == "Cannot handle two subfields with the same struct name but different structs"
    return
  raise Exception("Should've thrown AssertionError")

def test_const_connect_diffrent_structs_same_name():

  class A:
    @bitstruct
    class SomeMsg1:
      a: Bits8
      b: Bits32
  class B:
    @bitstruct
    class SomeMsg1:
      c: Bits8

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.out = OutPort(A.SomeMsg1)
      connect( s.out, A.SomeMsg1(1,3) )

      s.out2 = OutPort(B.SomeMsg1)
      connect( s.out2, B.SomeMsg1(4) )

  x = Top()
  x.elaborate()
  simple_sim_pass(x)
  x.tick()
  assert x.out == A.SomeMsg1(1,3)
  assert x.out2 == B.SomeMsg1(4)

def test_invalid_connect_outside_hierarchy():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = InPort(Bits32)

  try:
    x = Top()
    x.elaborate()
    connect( x.wire, Bits32(8) )
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

globalvar = 2
def test_connect_lambda():

  class Top( ComponentLevel3 ):
    def construct( s, x ):
      s.in_ = InPort(Bits32)
      s.out = OutPort(Bits32)

      s.out //= lambda: s.in_ + x + globalvar

  x = Top(3)
  x.elaborate()
  simple_sim_pass(x)
  x.in_ = 10
  x.tick()
  assert x.out == 10 + 3 + 2

  y = Top(33)
  y.elaborate()
  simple_sim_pass(y)
  y.in_ = 100
  y.tick()
  assert y.out == 100 + 33 + 2

def test_lambda_name_conflict():

  class Top( ComponentLevel3 ):
    def construct( s, x ):
      s.in_ = InPort(Bits32)
      s.out = OutPort(Bits32)
      s.out2 = OutPort(Bits32)

      s.out //= lambda: s.in_ + x

      # TODO throw some better error message when
      # the implicit name of a lambda function conflicts
      # with the explicit name of an update block

      @s.update
      def _lambda__s_out():
        s.out2 = Bits32(2)

  try:
    x = Top(3)
    x.elaborate()
  except UpblkFuncSameNameError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown UpblkFuncSameNameError.")

def test_invalid_in_out_loopback_at_self():

  class Comp( ComponentLevel3 ):
    def construct( s ):
      s.y = OutPort( Bits32 )
      s.z = OutPort( Bits32 )
      s.x = InPort( Bits32 )

      s.y //= Bits32(1)
      s.x //= s.y
      s.z //= s.x

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.comp = Comp()

  try:
    a = Top()
    a.elaborate()
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

def test_in_out_loopback_at_parent():

  class Comp( ComponentLevel3 ):
    def construct( s ):
      s.y = OutPort( Bits32 )
      s.z = OutPort( Bits32 )
      s.x = InPort( Bits32 )

      s.y //= Bits32(1)
      s.z //= s.x

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.comp = Comp()
      s.comp.x //= s.comp.y

  a = Top()
  a.elaborate()

def test_connect_slice_int():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.y = OutPort( Bits8 )
      s.x = Wire( Bits32 )

      s.y //= s.x[0:8]
      @s.update
      def sx():
        s.x = 10 # Except

  a = Top()
  a.elaborate()
  simple_sim_pass( a )
  try:
    a.tick() # expect to get int error
  except TypeError as e:
    assert str(e).startswith( "'int' object is not subscriptable" )
    return
  raise Exception("Should've thrown TypeError: 'int' object is not subscriptable")

def test_misconnect_component_to_signal():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.y = OutPort( Bits8 )
      s.x = Wire( Bits32 )
      s.y //= s # this is garbage

  a = Top()
  try:
    a.elaborate()
  except InvalidConnectionError as e:
    print(e)
    return
  raise Exception("Should've thrown InvalidConnectionError")

def test_s_connect_deprecated():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.y = OutPort( Bits8 )
      s.x = Wire( Bits32 )
      s.connect(s.x, s.y)

  a = Top()
  try:
    a.elaborate()
  except PyMTLDeprecationError as e:
    print(e)
    return
  raise Exception("Should've thrown InvalidConnectionError")
